# -*- coding: utf-8 -*-

"""Top-level package for biocircuits"""

__all__ = ["viz", "utils", "reg", "dynsys", "gillespie", "rd"]

from .viz import *
from .utils import *
from .reg import *
from .dynsys import *
from .gillespie import *
from .rd import *


__author__ = """Justin Bois"""
__email__ = "bois@caltech.edu"
__version__ = "0.1.4"



import warnings

import numpy as np

import matplotlib.streamplot

import bokeh.application
import bokeh.application.handlers
import bokeh.layouts
import bokeh.models
import bokeh.palettes
import bokeh.plotting

import colorcet

from . import utils


def _ecdf_vals(data, formal=False, complementary=False):
    """Get x, y, values of an ECDF for plotting.
    Parameters
    ----------
    data : ndarray
        One dimensional Numpy array with data.
    formal : bool, default False
        If True, generate x and y values for formal ECDF (staircase). If
        False, generate x and y values for ECDF as dots.
    complementary : bool
        If True, return values for ECCDF.
    Returns
    -------
    x : ndarray
        x-values for plot
    y : ndarray
        y-values for plot
    """
    x = np.sort(data)
    y = np.arange(1, len(data) + 1) / len(data)

    if formal:
        x, y = _to_formal(x, y)
        if complementary:
            y = 1 - y
    elif complementary:
        y = 1 - y + 1 / len(y)

    return x, y


def _to_formal(x, y):
    """Convert to formal ECDF."""
    # Set up output arrays
    x_formal = np.empty(2 * len(x))
    y_formal = np.empty(2 * len(x))

    # y-values for steps
    y_formal[0] = 0
    y_formal[1::2] = y
    y_formal[2::2] = y[:-1]

    # x- values for steps
    x_formal[::2] = x
    x_formal[1::2] = x

    return x_formal, y_formal


def ecdf(
    data=None,
    conf_int=False,
    ptiles=[2.5, 97.5],
    n_bs_reps=1000,
    fill_color="lightgray",
    fill_alpha=1,
    p=None,
    x_axis_label=None,
    y_axis_label=None,
    title=None,
    plot_height=300,
    plot_width=450,
    formal=False,
    complementary=False,
    x_axis_type="linear",
    y_axis_type="linear",
    **kwargs,
):
    """
    Create a plot of an ECDF. DEPRECATED. Use iqplot.ecdf().

    Parameters
    ----------
    data : array_like
        One-dimensional array of data. Nan's are ignored.
    conf_int : bool, default False
        If True, display a confidence interval on the ECDF.
    ptiles : list, default [2.5, 97.5]
        The percentiles to use for the confidence interval. Ignored it
        `conf_int` is False.
    n_bs_reps : int, default 1000
        Number of bootstrap replicates to do to compute confidence
        interval. Ignored if `conf_int` is False.
    fill_color : str, default 'lightgray'
        Color of the confidence interbal. Ignored if `conf_int` is
        False.
    fill_alpha : float, default 1
        Opacity of confidence interval. Ignored if `conf_int` is False.
    p : bokeh.plotting.Figure instance, or None (default)
        If None, create a new figure. Otherwise, populate the existing
        figure `p`.
    x_axis_label : str, default None
        Label for the x-axis. Ignored if `p` is not None.
    y_axis_label : str, default 'ECDF' or 'ECCDF'
        Label for the y-axis. Ignored if `p` is not None.
    title : str, default None
        Title of the plot. Ignored if `p` is not None.
    plot_height : int, default 300
        Height of plot, in pixels. Ignored if `p` is not None.
    plot_width : int, default 450
        Width of plot, in pixels. Ignored if `p` is not None.
    formal : bool, default False
        If True, make a plot of a formal ECDF (staircase). If False,
        plot the ECDF as dots.
    complementary : bool, default False
        If True, plot the empirical complementary cumulative
        distribution functon.
    x_axis_type : str, default 'linear'
        Either 'linear' or 'log'.
    y_axis_type : str, default 'linear'
        Either 'linear' or 'log'.
    kwargs
        Any kwargs to be passed to either p.circle or p.line, for
        `formal` being False or True, respectively.
    Returns
    -------
    output : bokeh.plotting.Figure instance
        Plot populated with ECDF.
    """
    warnings.warn(
        "`biocircuits.ecdf()` is deprecated. Used `iqplot.ecdf()` instead.",
        DeprecationWarning,
    )

    # Check data to make sure legit
    data = utils._convert_data(data)

    # Get y-axis label
    if p is None and y_axis_label is None:
        if complementary:
            y_axis_label = "ECCDF"
        else:
            y_axis_label = "ECDF"

    # Data points on ECDF
    x, y = _ecdf_vals(data, formal, complementary)

    # Instantiate Bokeh plot if not already passed in
    if p is None:
        p = bokeh.plotting.figure(
            plot_height=plot_height,
            plot_width=plot_width,
            x_axis_label=x_axis_label,
            y_axis_label=y_axis_label,
            x_axis_type=x_axis_type,
            y_axis_type=y_axis_type,
            title=title,
        )

    # Do bootstrap replicates
    if conf_int:
        x_plot = np.sort(np.unique(x))
        bs_reps = np.array(
            [
                _ecdf_arbitrary_points(np.random.choice(data, size=len(data)), x_plot)
                for _ in range(n_bs_reps)
            ]
        )

        # Compute the confidence intervals
        ecdf_low, ecdf_high = np.percentile(np.array(bs_reps), ptiles, axis=0)

        # Make them formal
        _, ecdf_low = _to_formal(x=x_plot, y=ecdf_low)
        x_plot, ecdf_high = _to_formal(x=x_plot, y=ecdf_high)

        p = fill_between(
            x1=x_plot,
            y1=ecdf_low,
            x2=x_plot,
            y2=ecdf_high,
            fill_color=fill_color,
            show_line=False,
            p=p,
        )

    if formal:
        # Line of steps
        p.line(x, y, **kwargs)

        # Rays for ends
        if complementary:
            p.ray(x[0], 1, None, np.pi, **kwargs)
            p.ray(x[-1], 0, None, 0, **kwargs)
        else:
            p.ray(x[0], 0, None, np.pi, **kwargs)
            p.ray(x[-1], 1, None, 0, **kwargs)
    else:
        p.circle(x, y, **kwargs)

    return p


def interactive_xy_plot(
    base_plot, callback, slider_params=(), toggle_params=(), extra_args=()
):
    """
    Create an interactive x-y plot in Bokeh. DEPRECATED.

    Parameters
    ----------
    base_plot : function
        A function to generate the initial plot that will be
        interactive. It must have call signature
        `base_plot(callback, sliders, toggles, extra_args))`, with the
        following arguments.
          callback:   A function to update the data source of the plot,
                      described as the `callback` argument below.
          sliders:    A tuple of `bokeh.models.Slider` objects.
                      Alternatively, can be any object reference-able
                      like `sliders[0].start`, `sliders[0].value`, etc.
          toggles:    A tuple of `bokeh.models.Toggle` objects.
                      Alternatively, can be any object reference-able
                      like `toggles[0].active`.
          extra_args: Tuple of any extra arguments that are passed to
                      the `callback` function.
        The `base_plot` function must return a Bokeh Figure instance and
        a Bokeh ColumnDataSource.
    callback : function
        A function that is executed to update the `ColumnDataSource` of
        the interactive plot whenever a slider or toggle are updated.
        It must have a call signature
        `callback(source, x_range, y_range, sliders, toggles,
                  *extra_args)`.
        Here, `source` is a `ColumnDataSource`, `x_range` is the
        x_range of the plot, and `y_range` is the y_range for the plot.
        `sliders`, `toggles`, and `extra_args` are as defined above.
    slider_params : tuple of objects, default empty tuple
        Each object in the tuple is an instance of a class that has the
        following attributes.
            title : The name of the slider.
            start : The smallest value of the slider.
            end : The largest value of the slider.
            value : The starting value of the slider.
            step : The step size of the slider as it is moved.
    toggle_params : tuple of objects, default empty tuple
        Each object in the tuple is an instance of a class that has the
        following attributes.
            title : The name of the toggle.
            active : A Boolean saying whether the toggle is active.
    extra_args : tuple, default empty tuple
        Tuple of any extra arguments that are passed to the
        `callback` function.

    Returns
    -------
    output : Bokeh application
        A Bokeh application with sliders, toggles, and a plot.
    """
    warnings.warn(
        DeprecationWarning,
        "`interactive_xy_plot() is deprecated. Either custom build an interactive plot with base Bokeh or use Panel.",
    )

    def _plot_app(doc):
        # Build the initial plot and data source
        p, source = base_plot(callback, slider_params, toggle_params, extra_args)

        # Make sure axis ranges have no padding
        if type(p.x_range) == bokeh.models.ranges.Range1d:
            start, end = p.x_range.start, p.x_range.end
            p.x_range = bokeh.models.ranges.DataRange1d(p.x_range)
            p.x_range.start = start
            p.x_range.end = end
        if type(p.y_range) == bokeh.models.ranges.Range1d:
            start, end = p.y_range.start, p.y_range.end
            p.y_range = bokeh.models.ranges.DataRange1d(p.y_range)
            p.y_range.start = start
            p.y_range.end = end
        p.x_range.range_padding = 0
        p.y_range.range_padding = 0

        # Callbacks
        def _callback(attr, old, new):
            callback(source, p.x_range, p.y_range, sliders, toggles, *extra_args)

        # Callback for the toggle with required call signature
        def _callback_toggle(new):
            _callback(None, None, new)

        # Set up sliders
        sliders = tuple(
            bokeh.models.Slider(
                start=param.start,
                end=param.end,
                value=param.value,
                step=param.step,
                title=param.title,
            )
            for param in slider_params
        )
        for slider in sliders:
            slider.on_change("value", _callback)

        # Set up toggles
        toggles = tuple(
            bokeh.models.Toggle(label=param.title) for param in toggle_params
        )
        for toggle in toggles:
            toggle.on_click(_callback_toggle)

        # Execute callback upon changing axis values
        p.x_range.on_change("start", _callback)
        p.x_range.on_change("end", _callback)
        p.y_range.on_change("start", _callback)
        p.y_range.on_change("end", _callback)

        # Add the plot to the app
        widgets = bokeh.layouts.column(*sliders, *toggles)
        doc.add_root(bokeh.layouts.column(widgets, p))

    handler = bokeh.application.handlers.FunctionHandler(_plot_app)
    return bokeh.application.Application(handler)


def xyt_plot(
    x,
    y,
    t,
    glyph="line",
    legend_names=None,
    legend_location="top_right",
    palette=None,
    glyph_kwargs={},
    time_slider_title="time",
    **kwargs,
):
    """Create an interactive plot some the output of `rd_solve()`.

    Parameters
    ----------
    x: 1D Numpy array
        x-values for plot.
    y : tuple of 2D Numpy arrays
        y-values for plot. `y[i][j, k]` is the y-value for curve i at
        time point j and position x[k].
    t : Numpy array
        y-values for plot.
    glyph : str, either "line" or "circle"
        What kind of glyph to use in plot.
    legend_names : list or None, default None
        Names for legend in plot.
    legend_location : str, default 'top_right'
        Position of legend using Bokeh's specifications for legend
        position. Ignored if `legend_names` is None.
    palette : List of colors default bokeh.palettes.d3['Category10'][10]
        Color palette to use for curves.
    kwargs :
        All other kwargs are passed to bokeh.plotting.figure() in
        creating the figure.

    Returns
    -------
    output
        An interactive plot with a time slider.
    """
    if palette is None:
        palette = colorcet.b_glasbey_category10

    if glyph not in ["line", "circle"]:
        raise RuntimeError("Only 'line' or 'circle' glyphs allowed.")

    # Set defaults
    if glyph == "line" and "line_width" not in glyph_kwargs:
        glyph_kwargs["line_width"] = 2
    if "x_axis_label" not in kwargs:
        kwargs["x_axis_label"] = "x"
    if "y_axis_label" not in kwargs:
        kwargs["y_axis_label"] = "y"
    if "frame_height" not in kwargs and "plot_height" not in kwargs:
        kwargs["frame_height"] = 275
    if "frame_width" not in kwargs and "plot_width" not in kwargs:
        kwargs["frame_width"] = 450
    if "y_range" not in kwargs:
        total_range_min = np.concatenate(y).min()
        total_range_max = np.concatenate(y).max()
        total_range = total_range_max - total_range_min
        kwargs["y_range"] = [
            total_range_min - total_range * 0.02,
            total_range_max + total_range * 0.02,
        ]
    if "x_range" not in kwargs:
        kwargs["x_range"] = [x.min(), x.max()]

    if "color" not in glyph_kwargs and len(y) > len(palette):
        raise RuntimeError("Not enough colors in palette for curves.")

    if legend_names is None:
        legend_names = [None for _ in range(len(y))]
    elif len(legend_names) != len(y):
        raise RuntimeError("len(legend_names) must equal len(y).")

    t_slider = bokeh.models.Slider(
        title=time_slider_title,
        name="t",
        start=t[0],
        end=t[-1],
        value=t[0],
        step=np.diff(t).mean(),
    )

    # Build data sources, source has all of the data
    data_dict = {"y_" + str(j): [] for j in range(len(y))}
    data_dict["i"] = []
    for i in range(len(t)):
        data_dict["i"] += [i] * len(x)
        for j, c in enumerate(y):
            data_dict["y_" + str(j)] += list(c[i, :])
    source = bokeh.models.ColumnDataSource(data_dict)

    # source_plot has only the data that will be displayed at a given time
    plot_dict = {
        "y_" + str(j): data_dict["y_" + str(j)][: len(x)] for j in range(len(y))
    }
    plot_dict["x"] = x
    source_plot = bokeh.models.ColumnDataSource(plot_dict)

    # The time points, need to have around
    source_t = bokeh.models.ColumnDataSource(dict(t=t))

    # Make the plot
    p = bokeh.plotting.figure(**kwargs)

    click_policy = False
    args = {key: val for key, val in glyph_kwargs.items()}
    args["source"] = source_plot
    for i, leg in enumerate(legend_names):
        if "color" not in glyph_kwargs:
            args["color"] = palette[i]

        if leg is not None:
            args["legend_label"] = str(leg)
            click_policy = True

        if glyph == "line":
            p.line("x", "y_" + str(i), **args)
        elif glyph == "circle":
            p.circle("x", "y_" + str(i), **args)

    if click_policy:
        p.legend.click_policy = "hide"
        p.legend.location = legend_location

    # Callback
    js_code = """
function sortedIndex(array, value) {
    var low = 0,
        high = array.length;

    while (low < high) {
        var mid = (low + high) >>> 1;
        if (array[mid] < value) low = mid + 1;
        else high = mid;
    }
    return low;
}

var x = source_plot.data['x'];
var x_len = x.length;
var t = source_t.data['t'];

var i = sortedIndex(t, cb_obj.value);

var k;
"""

    for var_name in ["y_" + str(j) for j in range(len(y))]:
        js_code += f"""var {var_name} = source_plot.data['{var_name}'];
var {var_name}_source = source.data['{var_name}'];
for (k = 0; k < x_len; k++) {var_name}[k] = {var_name}_source[x_len * i + k];

"""

    js_code += "source_plot.change.emit();\n"

    callback = bokeh.models.CustomJS(
        args=dict(source=source, source_t=source_t, source_plot=source_plot),
        code=js_code,
    )

    t_slider.js_on_change("value", callback)

    return bokeh.layouts.column(t_slider, p)


def phase_portrait(
    du_dt,
    dv_dt,
    u_range,
    v_range,
    args_u=(),
    args_v=(),
    log=False,
    p=None,
    zoomable=False,
    **kwargs,
):
    """
    Plots the phase portrait for a 2D dynamical system in the u-v plane.

    Parameters
    ----------
    du_dt : function
        A function to compute the right hand side of du/dt. Must have
        call signature `du_dt(u, v, *args_u)`. Note that there can be
        no explicit time dependence.
    dv_dt : function
        A function to compute the right hand side of dv/dt. Must have
        call signature `dv_dt(u, v, *args_v)`. Note that there can be
        no explicit time dependence.
    u_range : array_like, shape (2,)
        Minimum and maximum values of u to consider.
    v_range : array_like, shape (2,)
        Minimum and maximum values of v to consider.
    args_u : tuple, default ()
        Tuple of extra arguments to be passed to `du_dt`.
    args_v : tuple, default ()
        Tuple of extra arguments to be passed to `dv_dt`.
    log : bool, default False
        If True, plot u and v on a logarithmic scale.
    p : bokeh.plotting.Figure instance, default None
        Figure to use for the phase portrait. If None, a new one is
        created according to `streamplot()`.
    zoomable : bool, default False
        If True, generates a zoomable plot where the streamlines will be
        redrawn to correspond for the level so zoom. This requires
        Python to be running for the plot to render.
    kwargs :
        All other kwargs are passed to `streamplot`.

    Returns
    -------
    output : bokeh.plotting.Figure instance populated with streamplot
    """
    if zoomable:
        _kwargs = {f"_{key}": val for key, val in kwargs.items()}
        return _zoomable_phase_portrait(
            du_dt, dv_dt, u_range, v_range, args_u, args_v, log, p, **_kwargs
        )
    else:
        return _phase_portrait(
            du_dt, dv_dt, u_range, v_range, args_u, args_v, log, p, **kwargs
        )


def _phase_portrait(du_dt, dv_dt, u_range, v_range, args_u, args_v, log, p, **kwargs):
    if log:
        # Set up u,v space
        log_u = np.linspace(np.log10(u_range[0]), np.log10(u_range[1]), 100)
        log_v = np.linspace(np.log10(v_range[0]), np.log10(v_range[1]), 100)
        log_uu, log_vv = np.meshgrid(log_u, log_v)

        # Compute derivatives
        log_u_vel = du_dt(10 ** log_uu, 10 ** log_vv, *args_u) / 10 ** log_uu
        log_v_vel = dv_dt(10 ** log_uu, 10 ** log_vv, *args_v) / 10 ** log_vv

        # Make stream plot
        return streamplot(log_u, log_v, log_u_vel, log_v_vel, p=p, **kwargs)
    else:
        # Set up u,v space
        u = np.linspace(u_range[0], u_range[1], 100)
        v = np.linspace(v_range[0], v_range[1], 100)
        uu, vv = np.meshgrid(u, v)

        # Compute derivatives
        u_vel = du_dt(uu, vv, *args_u)
        v_vel = dv_dt(uu, vv, *args_v)

        # Make stream plot
        return streamplot(u, v, u_vel, v_vel, p=p, **kwargs)


def streamplot(
    x,
    y,
    u,
    v,
    p=None,
    density=1,
    color="#1f77b4",
    line_width=None,
    alpha=1,
    arrow_size=7,
    min_length=0.1,
    start_points=None,
    max_length=4.0,
    integration_direction="both",
    arrow_level="underlay",
    **kwargs,
):
    """Draws streamlines of a vector field.

    Parameters
    ----------
    x, y : 1d arrays
        an evenly spaced grid.
    u, v : 2d arrays
        x and y-velocities. Number of rows should match length of y, and
        the number of columns should match x.
    p : bokeh.plotting.Figure instance, default None
        Figure to populate with glyphs. If None, create a new figure.
    density : float or 2-tuple
        Controls the closeness of streamlines. When `density = 1`, the domain
        is divided into a 30x30 grid---density linearly scales this grid.
        Each cell in the grid can have, at most, one traversing streamline.
        For different densities in each direction, use [density_x, density_y].
    color : str or 2d array, default '#1f77b4' (Bokeh default color)
        Streamline color. When given an array with the same shape as
        velocities, color values are converted to colors using cmap.
    line_width : numeric or 2d array, default None
        vary linewidth when given a 2d array with the same shape as velocities. If None, scale linewidth with speed.
    arrow_size : float
        Factor scale arrow size.
    min_length : float
        Minimum length of streamline in axes coordinates.
    start_points: Nx2 array
        Coordinates of starting points for the streamlines.
        In data coordinates, the same as the ``x`` and ``y`` arrays.
    max_length : float
        Maximum length of streamline in axes coordinates.
    integration_direction : ['forward', 'backward', 'both']
        Integrate the streamline in forward, backward or both directions.
    arrow_level : str
        Either 'underlay' or 'overlay'.
    kwargs :
        All other kwargs are passed to bokeh.plotting.figure() when
        generating the figure.

    Returns
    -------
    bokeh.plotting.Figure instance populated with streamplot.

    Notes
    -----
    .. Adapted from matplotlib.streamplot.streamplot.py.
    """
    if p is None:
        p = _baseplot(p, **kwargs)

    # Ensure plot fits stream lines
    p.x_range = bokeh.models.Range1d(x[0], x[-1])
    p.y_range = bokeh.models.Range1d(y[0], y[-1])

    if line_width is None:
        # Compute speed
        speed = np.sqrt(u ** 2 + v ** 2)

        # Make linewidths proportional to speed, with min width 0.5 and max 3
        line_width = 0.5 + 2.5 * speed / speed.max()

    xs, ys, line_widths, arrowtails, arrowheads = _streamlines(
        x,
        y,
        u,
        v,
        density=density,
        line_width=line_width,
        min_length=min_length,
        start_points=start_points,
        max_length=max_length,
        integration_direction=integration_direction,
    )

    def _draw_arrows():
        for tail, head in zip(arrowtails, arrowheads):
            p.add_layout(
                bokeh.models.Arrow(
                    line_alpha=0,
                    end=bokeh.models.NormalHead(fill_color=color, line_alpha=0, size=7),
                    x_start=tail[0],
                    y_start=tail[1],
                    x_end=head[0],
                    y_end=head[1],
                )
            )

    if arrow_level == "underlay":
        _draw_arrows()
        p.multi_line(xs, ys, color=color, line_width=line_widths, line_alpha=alpha)
    else:
        p.multi_line(xs, ys, color=color, line_width=line_widths, line_alpha=alpha)
        _draw_arrows()

    return p


def _baseplot(p, **kwargs):
    if p is None:
        if "x_axis_label" not in kwargs:
            kwargs["x_axis_label"] = "x"
        if "y_axis_label" not in kwargs:
            kwargs["y_axis_label"] = "y"
        if "x_range" in kwargs or "y_range" in kwargs:
            raise RuntimeError("Cannot specify ranges in kwargs.")
        if (
            "plot_height" not in kwargs
            or "frame_height" not in kwargs
            or "height" not in kwargs
        ):
            kwargs["frame_height"] = 260
        if (
            "plot_width" not in kwargs
            or "frame_width" not in kwargs
            or "width" not in kwargs
        ):
            kwargs["frame_width"] = 260
        if "tools" not in kwargs:
            tools = "box_zoom,save,reset"

        p = bokeh.plotting.figure(**kwargs)

    return p


def _give_back_local_params(
    du_dt,
    dv_dt,
    u_range,
    v_range,
    args_u,
    args_v,
    log,
    p,
    density,
    color,
    line_width,
    alpha,
    arrow_size,
    min_length,
    start_points,
    max_length,
    integration_direction,
    arrow_level,
):

    return (
        du_dt,
        dv_dt,
        u_range,
        v_range,
        args_u,
        args_v,
        log,
        p,
        density,
        color,
        line_width,
        alpha,
        arrow_size,
        min_length,
        start_points,
        max_length,
        integration_direction,
        arrow_level,
    )


def _zoomable_phase_portrait(
    _du_dt,
    _dv_dt,
    _u_range,
    _v_range,
    _args_u,
    _args_v,
    _log,
    _p,
    _density=1,
    _color="#1f77b4",
    _line_width=None,
    _alpha=1,
    _arrow_size=7,
    _min_length=0.1,
    _start_points=None,
    _max_length=4.0,
    _integration_direction="both",
    _arrow_level="underlay",
    **kwargs,
):
    """Draws streamlines of a vector flow.

    Parameters
    ----------
    x, y : 1d arrays
        an evenly spaced grid.
    u, v : 2d arrays
        x and y-velocities. Number of rows should match length of y, and
        the number of columns should match x.
    p : bokeh.plotting.Figure instance, default None
        Figure to populate with glyphs. If None, create a new figure.
    density : float or 2-tuple
        Controls the closeness of streamlines. When `density = 1`, the domain
        is divided into a 30x30 grid---density linearly scales this grid.
        Each cell in the grid can have, at most, one traversing streamline.
        For different densities in each direction, use [density_x, density_y].
    color : str or 2d array, default '#1f77b4' (Bokeh default color)
        Streamline color. When given an array with the same shape as
        velocities, color values are converted to colors using cmap.
    line_width : numeric or 2d array, default None
        vary linewidth when given a 2d array with the same shape as velocities. If None, scale linewidth with speed.
    arrow_size : float
        Factor scale arrow size.
    min_length : float
        Minimum length of streamline in axes coordinates.
    start_points: Nx2 array
        Coordinates of starting points for the streamlines.
        In data coordinates, the same as the ``x`` and ``y`` arrays.
    max_length : float
        Maximum length of streamline in axes coordinates.
    integration_direction : ['forward', 'backward', 'both']
        Integrate the streamline in forward, backward or both directions.
    arrow_level : str
        Either 'underlay' or 'overlay'.
    kwargs :
        All other kwargs are passed to bokeh.plotting.figure() when
        generating the figure.

    Returns
    -------
    bokeh.plotting.Figure instance populated with streamplot.

    Notes
    -----
    .. Adapted from matplotlib.streamplot.streamplot.py.
    """

    def _plot_app(doc):
        (
            du_dt,
            dv_dt,
            u_range,
            v_range,
            args_u,
            args_v,
            log,
            p,
            density,
            color,
            line_width,
            alpha,
            arrow_size,
            min_length,
            start_points,
            max_length,
            integration_direction,
            arrow_level,
        ) = _give_back_local_params(
            _du_dt,
            _dv_dt,
            _u_range,
            _v_range,
            _args_u,
            _args_v,
            _log,
            _p,
            _density,
            _color,
            _line_width,
            _alpha,
            _arrow_size,
            _min_length,
            _start_points,
            _max_length,
            _integration_direction,
            _arrow_level,
        )
        p = _baseplot(p, **kwargs)

        # Ensure plot fits stream lines
        if log:
            p.x_range = bokeh.models.Range1d(
                np.log10(u_range[0]), np.log10(u_range[-1])
            )
            p.y_range = bokeh.models.Range1d(
                np.log10(v_range[0]), np.log10(v_range[-1])
            )
        else:
            p.x_range = bokeh.models.Range1d(u_range[0], u_range[-1])
            p.y_range = bokeh.models.Range1d(v_range[0], v_range[-1])

        # Make sure axis ranges have no padding
        if type(p.x_range) == bokeh.models.ranges.Range1d:
            start, end = p.x_range.start, p.x_range.end
            p.x_range = bokeh.models.ranges.DataRange1d(p.x_range)
            p.x_range.start = start
            p.x_range.end = end
        if type(p.y_range) == bokeh.models.ranges.Range1d:
            start, end = p.y_range.start, p.y_range.end
            p.y_range = bokeh.models.ranges.DataRange1d(p.y_range)
            p.y_range.start = start
            p.y_range.end = end
        p.x_range.range_padding = 0
        p.y_range.range_padding = 0

        # Only allow certain tools
        p.tools = [
            bokeh.models.BoxZoomTool(),
            bokeh.models.ResetTool(),
            bokeh.models.SaveTool(),
        ]

        if log:
            # Set up u,v space
            u = np.linspace(np.log10(u_range[0]), np.log10(u_range[1]), 100)
            v = np.linspace(np.log10(v_range[0]), np.log10(v_range[1]), 100)
            log_uu, log_vv = np.meshgrid(u, v)

            # Compute derivatives
            u_vel = du_dt(10 ** log_uu, 10 ** log_vv, *args_u) / 10 ** log_uu
            v_vel = dv_dt(10 ** log_uu, 10 ** log_vv, *args_v) / 10 ** log_vv
        else:
            # Set up u,v space
            u = np.linspace(u_range[0], u_range[1], 100)
            v = np.linspace(v_range[0], v_range[1], 100)
            uu, vv = np.meshgrid(u, v)

            # Compute derivatives
            u_vel = du_dt(uu, vv, *args_u)
            v_vel = dv_dt(uu, vv, *args_v)

        if line_width is None:
            # Compute speed
            speed = np.sqrt(u_vel ** 2 + v_vel ** 2)

            # Make linewidths proportional to speed, with min width 0.5 and max 3
            line_width_plot = 0.5 + 2.5 * speed / speed.max()
        else:
            line_width_plot = line_width

        xs, ys, line_widths, arrowtails, arrowheads = _streamlines(
            u,
            v,
            u_vel,
            v_vel,
            density=density,
            line_width=line_width_plot,
            min_length=min_length,
            start_points=start_points,
            max_length=max_length,
            integration_direction=integration_direction,
        )

        line_source = bokeh.models.ColumnDataSource(
            dict(xs=xs, ys=ys, line_width=line_widths)
        )

        arrows = [
            bokeh.models.Arrow(
                line_alpha=0,
                end=bokeh.models.NormalHead(fill_color=color, line_alpha=0, size=7),
                x_start=tail[0],
                y_start=tail[1],
                x_end=head[0],
                y_end=head[1],
            )
            for head, tail in zip(arrowheads, arrowtails)
        ]

        if arrow_level == "underlay":
            for arrow in arrows:
                p.add_layout(arrow)

            p.multi_line(
                source=line_source,
                xs="xs",
                ys="ys",
                line_color=color,
                line_width="line_width",
                line_alpha=alpha,
            )
        else:
            p.multi_line(
                source=line_source,
                xs="xs",
                ys="ys",
                line_color=color,
                line_width="line_width",
                line_alpha=alpha,
            )

            for arrow in arrows:
                p.add_layout(arrow)

        def _callback(attr, old, new):
            # Set up u,v space
            u = np.linspace(p.x_range.start, p.x_range.end, 100)
            v = np.linspace(p.y_range.start, p.y_range.end, 100)
            uu, vv = np.meshgrid(u, v)

            if log:
                # Compute derivatives
                u_vel = du_dt(10 ** uu, 10 ** vv, *args_u) / 10 ** uu
                v_vel = dv_dt(10 ** uu, 10 ** vv, *args_v) / 10 ** vv
            else:
                # Compute derivatives
                u_vel = du_dt(uu, vv, *args_u)
                v_vel = dv_dt(uu, vv, *args_v)

            if line_width is None:
                # Compute speed
                speed = np.sqrt(u_vel ** 2 + v_vel ** 2)

                # Make linewidths proportional to speed, with min width 0.5 and max 3
                line_width_plot = 0.5 + 2.5 * speed / speed.max()
            else:
                line_width_plot = line_width

            xs, ys, line_widths, arrowtails, arrowheads = _streamlines(
                u,
                v,
                u_vel,
                v_vel,
                density=density,
                line_width=line_width_plot,
                min_length=min_length,
                start_points=start_points,
                max_length=max_length,
                integration_direction=integration_direction,
            )

            line_source.data = dict(xs=xs, ys=ys, line_width=line_widths)

            if len(arrowtails) < len(arrows):
                for i, (head, tail) in enumerate(zip(arrowheads, arrowtails)):
                    arrows[i].x_start = tail[0]
                    arrows[i].y_start = tail[1]
                    arrows[i].x_end = head[0]
                    arrows[i].y_end = head[1]
                for i in range(len(arrowtails), len(arrows)):
                    arrows[i].visible = False
            else:
                for i, (head, tail) in enumerate(
                    zip(arrowheads[: len(arrows)], arrowtails[: len(arrows)])
                ):
                    arrows[i].x_start = tail[0]
                    arrows[i].y_start = tail[1]
                    arrows[i].x_end = head[0]
                    arrows[i].y_end = head[1]
                    arrows[i].visible = True

                for head, tail in zip(
                    arrowheads[len(arrows) :], arrowtails[len(arrows) :]
                ):
                    arrows.append(
                        bokeh.models.Arrow(
                            line_alpha=0,
                            end=bokeh.models.NormalHead(
                                fill_color=color, line_alpha=0, size=7,
                            ),
                            x_start=tail[0],
                            y_start=tail[1],
                            x_end=head[0],
                            y_end=head[1],
                        )
                    )

                    p.add_layout(arrows[-1])

        # Execute callback upon changing axis values
        p.x_range.on_change("start", _callback)
        p.x_range.on_change("end", _callback)
        p.y_range.on_change("start", _callback)
        p.y_range.on_change("end", _callback)

        # Add the plot to the app
        doc.add_root(p)

    handler = bokeh.application.handlers.FunctionHandler(_plot_app)
    return bokeh.application.Application(handler)


def _streamlines(
    x,
    y,
    u,
    v,
    density=1,
    line_width=1,
    min_length=0.1,
    start_points=None,
    max_length=4.0,
    integration_direction="both",
):
    """Gives specs for streamlines of a vector flow."""
    grid = matplotlib.streamplot.Grid(x, y)
    mask = matplotlib.streamplot.StreamMask(density)
    dmap = matplotlib.streamplot.DomainMap(grid, mask)

    if integration_direction not in ["both", "forward", "backward"]:
        errstr = (
            "Integration direction '%s' not recognised. "
            "Expected 'both', 'forward' or 'backward'." % integration_direction
        )
        raise ValueError(errstr)

    if integration_direction == "both":
        max_length /= 2.0

    if isinstance(line_width, np.ndarray):
        if line_width.shape != grid.shape:
            raise ValueError(
                "If 'line_width' is given, must have the shape of 'Grid(x,y)'"
            )
        line_widths = []
    else:
        line_widths = line_width

    ## Sanity checks.
    if u.shape != grid.shape or v.shape != grid.shape:
        raise ValueError("'u' and 'v' must be of shape 'Grid(x,y)'")

    u = np.ma.masked_invalid(u)
    v = np.ma.masked_invalid(v)

    integrate = matplotlib.streamplot.get_integrator(
        u, v, dmap, min_length, max_length, integration_direction
    )

    trajectories = []
    if start_points is None:
        for xm, ym in matplotlib.streamplot._gen_starting_points(mask.shape):
            if mask[ym, xm] == 0:
                xg, yg = dmap.mask2grid(xm, ym)
                t = integrate(xg, yg)
                if t is not None:
                    trajectories.append(t)
    else:
        sp2 = np.asanyarray(start_points, dtype=float).copy()

        # Check if start_points are outside the data boundaries
        for xs, ys in sp2:
            if not (
                grid.x_origin <= xs <= grid.x_origin + grid.width
                and grid.y_origin <= ys <= grid.y_origin + grid.height
            ):
                raise ValueError(
                    "Starting point ({}, {}) outside of data "
                    "boundaries".format(xs, ys)
                )

        # Convert start_points from data to array coords
        # Shift the seed points from the bottom left of the data so that
        # data2grid works properly.
        sp2[:, 0] -= grid.x_origin
        sp2[:, 1] -= grid.y_origin

        for xs, ys in sp2:
            xg, yg = dmap.data2grid(xs, ys)
            t = integrate(xg, yg)
            if t is not None:
                trajectories.append(t)

    slines = []
    arrowtails = []
    arrowheads = []
    for t in trajectories:
        tgx = np.array(t[0])
        tgy = np.array(t[1])
        # Rescale from grid-coordinates to data-coordinates.
        tx, ty = dmap.grid2data(*np.array(t))
        tx += grid.x_origin
        ty += grid.y_origin

        points = np.transpose([tx, ty]).reshape(-1, 1, 2)
        slines.extend(np.hstack([points[:-1], points[1:]]))

        # Add arrows half way along each trajectory.
        s = np.cumsum(np.sqrt(np.diff(tx) ** 2 + np.diff(ty) ** 2))
        n = np.searchsorted(s, s[-1] / 2.0)
        arrowtails.append((tx[n], ty[n]))
        arrowheads.append((np.mean(tx[n : n + 2]), np.mean(ty[n : n + 2])))

        if isinstance(line_width, np.ndarray):
            line_widths.extend(
                matplotlib.streamplot.interpgrid(line_width, tgx, tgy)[:-1]
            )

    xs = [s[:, 0] for s in slines]
    ys = [s[:, 1] for s in slines]

    return xs, ys, line_widths, arrowtails, arrowheads


def imshow(
    im,
    color_mapper=None,
    frame_height=400,
    frame_width=None,
    length_units="pixels",
    interpixel_distance=1.0,
    x_range=None,
    y_range=None,
    colorbar=False,
    no_ticks=False,
    x_axis_label=None,
    y_axis_label=None,
    title=None,
    flip=True,
    return_im=False,
    saturate_channels=True,
    min_intensity=None,
    max_intensity=None,
    display_clicks=False,
):
    """
    Display an image in a Bokeh figure.

    Parameters
    ----------
    im : Numpy array
        If 2D, intensity image to be displayed. If 3D, first two
        dimensions are pixel values. Last dimension can be of length
        1, 2, or 3, which specify colors.
    color_mapper : str or bokeh.models.LinearColorMapper, default None
        If `im` is an intensity image, `color_mapper` is a mapping of
        intensity to color. If None, default is 256-level Viridis.
        If `im` is a color image, then `color_mapper` can either be
        'rgb' or 'cmy' (default), for RGB or CMY merge of channels.
    plot_height : int
        Height of the plot in pixels. The width is scaled so that the
        x and y distance between pixels is the same.
    plot_width : int or None (default)
        If None, the width is scaled so that the x and y distance
        between pixels is approximately the same. Otherwise, the width
        of the plot in pixels.
    length_units : str, default 'pixels'
        The units of length in the image.
    interpixel_distance : float, default 1.0
        Interpixel distance in units of `length_units`.
    x_range : bokeh.models.Range1d instance, default None
        Range of x-axis. If None, determined automatically.
    y_range : bokeh.models.Range1d instance, default None
        Range of y-axis. If None, determined automatically.
    colorbar : bool, default False
        If True, include a colorbar.
    no_ticks : bool, default False
        If True, no ticks are displayed. See note below.
    x_axis_label : str, default None
        Label for the x-axis. If None, labeled with `length_units`.
    y_axis_label : str, default None
        Label for the y-axis. If None, labeled with `length_units`.
    title : str, default None
        The title of the plot.
    flip : bool, default True
        If True, flip image so it displays right-side up. This is
        necessary because traditionally images have their 0,0 pixel
        index in the top left corner, and not the bottom left corner.
    return_im : bool, default False
        If True, return the GlyphRenderer instance of the image being
        displayed.
    saturate_channels : bool, default True
        If True, each of the channels have their displayed pixel values
        extended to range from 0 to 255 to show the full dynamic range.
    min_intensity : int or float, default None
        Minimum possible intensity of a pixel in the image. If None,
        the image is scaled based on the dynamic range in the image.
    max_intensity : int or float, default None
        Maximum possible intensity of a pixel in the image. If None,
        the image is scaled based on the dynamic range in the image.
    display_clicks : bool, default False
        If True, display clicks to the right of the plot using
        JavaScript. The clicks are not recorded nor stored, just
        printed.

    Returns
    -------
    p : bokeh.plotting.figure instance
        Bokeh plot with image displayed.
    im : bokeh.models.renderers.GlyphRenderer instance (optional)
        The GlyphRenderer instance of the image being displayed. This is
        only returned if `return_im` is True.
    """

    # If a single channel in 3D image, flatten and check shape
    if im.ndim == 3:
        if im.shape[2] == 1:
            im = im[:, :, 0]
        elif im.shape[2] not in [2, 3]:
            raise RuntimeError("Can only display 1, 2, or 3 channels.")

    # If binary image, make sure it's int
    if im.dtype == bool:
        im = im.astype(np.uint8)

    # Get color mapper
    if im.ndim == 2:
        if color_mapper is None:
            color_mapper = bokeh.models.LinearColorMapper(bokeh.palettes.viridis(256))
        elif type(color_mapper) == str and color_mapper.lower() in ["rgb", "cmy"]:
            raise RuntimeError("Cannot use rgb or cmy colormap for intensity image.")
        if min_intensity is None:
            color_mapper.low = im.min()
        else:
            color_mapper.low = min_intensity
        if max_intensity is None:
            color_mapper.high = im.max()
        else:
            color_mapper.high = max_intensity
    elif im.ndim == 3:
        if color_mapper is None or color_mapper.lower() == "cmy":
            im = im_merge(
                *np.rollaxis(im, 2),
                cmy=True,
                im_0_min=min_intensity,
                im_1_min=min_intensity,
                im_2_min=min_intensity,
                im_0_max=max_intensity,
                im_1_max=max_intensity,
                im_2_max=max_intensity,
            )
        elif color_mapper.lower() == "rgb":
            im = im_merge(
                *np.rollaxis(im, 2),
                cmy=False,
                im_0_min=min_intensity,
                im_1_min=min_intensity,
                im_2_min=min_intensity,
                im_0_max=max_intensity,
                im_1_max=max_intensity,
                im_2_max=max_intensity,
            )
        else:
            raise RuntimeError("Invalid color mapper for color image.")
    else:
        raise RuntimeError("Input image array must have either 2 or 3 dimensions.")

    # Get shape, dimensions
    n, m = im.shape[:2]
    if x_range is not None and y_range is not None:
        dw = x_range[1] - x_range[0]
        dh = y_range[1] - y_range[0]
    else:
        dw = m * interpixel_distance
        dh = n * interpixel_distance
        x_range = [0, dw]
        y_range = [0, dh]

    # Set up figure with appropriate dimensions
    if frame_width is None:
        frame_width = int(m / n * frame_height)
    if colorbar:
        toolbar_location = "above"
    else:
        toolbar_location = "right"
    p = bokeh.plotting.figure(
        frame_height=frame_height,
        frame_width=frame_width,
        x_range=x_range,
        y_range=y_range,
        title=title,
        toolbar_location=toolbar_location,
        tools="pan,box_zoom,wheel_zoom,save,reset",
    )

    if no_ticks:
        p.xaxis.major_label_text_font_size = "0pt"
        p.yaxis.major_label_text_font_size = "0pt"
        p.xaxis.major_tick_line_color = None
        p.xaxis.minor_tick_line_color = None
        p.yaxis.major_tick_line_color = None
        p.yaxis.minor_tick_line_color = None
    else:
        if x_axis_label is None:
            p.xaxis.axis_label = length_units
        else:
            p.xaxis.axis_label = x_axis_label
        if y_axis_label is None:
            p.yaxis.axis_label = length_units
        else:
            p.yaxis.axis_label = y_axis_label

    # Display the image
    if im.ndim == 2:
        if flip:
            im = im[::-1, :]
        im_bokeh = p.image(
            image=[im],
            x=x_range[0],
            y=y_range[0],
            dw=dw,
            dh=dh,
            color_mapper=color_mapper,
        )
    else:
        im_bokeh = p.image_rgba(
            image=[rgb_to_rgba32(im, flip=flip)],
            x=x_range[0],
            y=y_range[0],
            dw=dw,
            dh=dh,
        )

    # Make a colorbar
    if colorbar:
        if im.ndim == 3:
            warnings.warn("No colorbar display for RGB images.")
        else:
            color_bar = bokeh.models.ColorBar(
                color_mapper=color_mapper,
                label_standoff=12,
                border_line_color=None,
                location=(0, 0),
            )
            p.add_layout(color_bar, "right")

    if display_clicks:
        div = bokeh.models.Div(width=200)
        layout = bokeh.layouts.row(p, div)
        p.js_on_event(bokeh.events.Tap, _display_clicks(div, attributes=["x", "y"]))
        if return_im:
            return layout, im_bokeh
        else:
            return layout

    if return_im:
        return p, im_bokeh
    return p


def _display_clicks(div, attributes=[], style="float:left;clear:left;font_size=0.5pt"):
    """Build a suitable CustomJS to display the current event
    in the div model."""
    return bokeh.models.CustomJS(
        args=dict(div=div),
        code="""
        var attrs = %s; var args = [];
        for (var i=0; i<attrs.length; i++ ) {
            args.push(Number(cb_obj[attrs[i]]).toFixed(4));
        }
        var line = "<span style=%r>[" + args.join(", ") + "], </span>\\n";
        var text = div.text.concat(line);
        var lines = text.split("\\n")
        if ( lines.length > 35 ) { lines.shift(); }
        div.text = lines.join("\\n");
    """
        % (attributes, style),
    )



def rep_hill(x, n):
    """Dimensionless production rate for a gene repressed by x.

    Parameters
    ----------
    x : float or NumPy array
        Concentration of repressor.
    n : float
        Hill coefficient.

    Returns
    -------
    output : NumPy array or float
        1 / (1 + x**n)
    """
    return 1.0 / (1.0 + x ** n)


def act_hill(x, n):
    """Dimensionless production rate for a gene activated by x.

    Parameters
    ----------
    x : float or NumPy array
        Concentration of activator.
    n : float
        Hill coefficient.

    Returns
    -------
    output : NumPy array or float
        x**n / (1 + x**n)
    """
    return 1.0 - rep_hill(x, n)


def aa_and(x, y, nx, ny):
    """Dimensionless production rate for a gene regulated by two
    activators with AND logic in the absence of leakage.

    Parameters
    ----------
    x : float or NumPy array
        Concentration of first activator.
    y : float or NumPy array
        Concentration of second activator.
    nx : float
        Hill coefficient for first activator.
    ny : float
        Hill coefficient for second activator.

    Returns
    -------
    output : NumPy array or float
        x**nx * y**ny / (1 + x**nx) / (1 + y**ny)
    """
    return x ** nx * y ** ny / (1.0 + x ** nx) / (1.0 + y ** ny)


def aa_or(x, y, nx, ny):
    """Dimensionless production rate for a gene regulated by two
    activators with OR logic in the absence of leakage.

    Parameters
    ----------
    x : float or NumPy array
        Concentration of first activator.
    y : float or NumPy array
        Concentration of second activator.
    nx : float
        Hill coefficient for first activator.
    ny : float
        Hill coefficient for second activator.

    Returns
    -------
    output : NumPy array or float
        (x**nx + y**ny + x**nx * y**ny) / (1 + x**nx) / (1 + y**ny)
    """
    denom = (1.0 + x ** nx) * (1.0 + y ** ny)
    return (denom - 1.0) / denom


def aa_or_single(x, y, nx, ny):
    """Dimensionless production rate for a gene regulated by two
    activators with OR logic in the absence of leakage with single
    occupancy.

    Parameters
    ----------
    x : float or NumPy array
        Concentration of first activator.
    y : float or NumPy array
        Concentration of second activator.
    nx : float
        Hill coefficient for first activator.
    ny : float
        Hill coefficient for second activator.

    Returns
    -------
    output : NumPy array or float
        (x**nx + y**ny) / (1 + x**nx + y**ny)
    """
    num = x ** nx + y ** ny
    return num / (1.0 + num)


def rr_and(x, y, nx, ny):
    """Dimensionless production rate for a gene regulated by two
    repressors with AND logic in the absence of leakage.

    Parameters
    ----------
    x : float or NumPy array
        Concentration of first repressor.
    y : float or NumPy array
        Concentration of second repressor.
    nx : float
        Hill coefficient for first repressor.
    ny : float
        Hill coefficient for second repressor.

    Returns
    -------
    output : NumPy array or float
        1 / (1 + x**nx) / (1 + y**ny)
    """
    return 1.0 / (1.0 + x ** nx) / (1.0 + y ** ny)


def rr_and_single(x, y, nx, ny):
    """Dimensionless production rate for a gene regulated by two
    repressors with AND logic in the absence of leakage with
    single occupancy.

    Parameters
    ----------
    x : float or NumPy array
        Concentration of first repressor.
    y : float or NumPy array
        Concentration of second repressor.
    nx : float
        Hill coefficient for first repressor.
    ny : float
        Hill coefficient for second repressor.

    Returns
    -------
    output : NumPy array or float
        1 / (1 + x**nx + y**ny)
    """
    return 1.0 / (1.0 + x ** nx + y ** ny)


def rr_or(x, y, nx, ny):
    """Dimensionless production rate for a gene regulated by two
    repressors with OR logic in the absence of leakage.

    Parameters
    ----------
    x : float or NumPy array
        Concentration of first repressor.
    y : float or NumPy array
        Concentration of second repressor.
    nx : float
        Hill coefficient for first repressor.
    ny : float
        Hill coefficient for second repressor.

    Returns
    -------
    output : NumPy array or float
        (1 + x**nx + y**ny) / (1 + x**nx) / (1 + y**ny)
    """
    return (1.0 + x ** nx + y ** ny) / (1.0 + x ** nx) / (1.0 + y ** ny)


def ar_and(x, y, nx, ny):
    """Dimensionless production rate for a gene regulated by one
    activator and one repressor with AND logic in the absence of
    leakage.

    Parameters
    ----------
    x : float or NumPy array
        Concentration of activator.
    y : float or NumPy array
        Concentration of repressor.
    nx : float
        Hill coefficient for activator.
    ny : float
        Hill coefficient for repressor.

    Returns
    -------
    output : NumPy array or float
        x ** nx / (1 + x**nx) / (1 + y**ny)
    """
    return x ** nx / (1.0 + x ** nx) / (1.0 + y ** ny)


def ar_or(x, y, nx, ny):
    """Dimensionless production rate for a gene regulated by one
    activator and one repressor with OR logic in the absence of
    leakage.

    Parameters
    ----------
    x : float or NumPy array
        Concentration of activator.
    y : float or NumPy array
        Concentration of repressor.
    nx : float
        Hill coefficient for activator.
    ny : float
        Hill coefficient for repressor.

    Returns
    -------
    output : NumPy array or float
        (1 + x**nx + x**nx * y**ny)) / (1 + x**nx) / (1 + y**ny)
    """
    return (1.0 + x ** nx * (1.0 + y ** ny)) / (1.0 + x ** nx) / (1.0 + y ** ny)


def ar_and_single(x, y, nx, ny):
    """Dimensionless production rate for a gene regulated by one
    activator and one repressor with AND logic in the absence of
    leakage with single occupancy.

    Parameters
    ----------
    x : float or NumPy array
        Concentration of activator.
    y : float or NumPy array
        Concentration of repressor.
    nx : float
        Hill coefficient for activator.
    ny : float
        Hill coefficient for repressor.

    Returns
    -------
    output : NumPy array or float
        x ** nx / (1 + x**nx + y**ny)
    """
    return x ** nx / (1.0 + x ** nx + y ** ny)


def ar_or_single(x, y, nx, ny):
    """Dimensionless production rate for a gene regulated by one
    activator and one repressor with OR logic in the absence of
    leakage with single occupancy.

    Parameters
    ----------
    x : float or NumPy array
        Concentration of activator.
    y : float or NumPy array
        Concentration of repressor.
    nx : float
        Hill coefficient for activator.
    ny : float
        Hill coefficient for repressor.

    Returns
    -------
    output : NumPy array or float
        (1 + x**nx) / (1 + x**nx + y**ny)
    """
    return (1.0 + x ** nx) / (1.0 + x ** nx + y ** ny)



import numpy as np
import scipy.integrate

from .. import reg

import bokeh.io
import bokeh.layouts
import bokeh.models
import bokeh.plotting

import colorcet


def _ffl_rhs(beta, gamma, kappa, n_xy, n_xz, n_yz, ffl, logic):
    """Return a function with call signature fun(yz, x) that computes
    the right-hand side of the dynamical system for an FFL. Here,
    `yz` is a length two array containing concentrations of Y and Z.
    """
    if ffl[:2].lower() in ("c1", "c3", "i1", "i3"):
        fy = lambda x: reg.act_hill(x, n_xy)
    else:
        fy = lambda x: reg.rep_hill(x, n_xy)

    if ffl[:2].lower() in ("c1", "i4"):
        if logic.lower() == "and":
            fz = lambda x, y: reg.aa_and(x, y, n_xz, n_yz)
        else:
            fz = lambda x, y: reg.aa_or(x, y, n_xz, n_yz)
    elif ffl[:2].lower() in ("c4", "i1"):
        if logic.lower() == "and":
            fz = lambda x, y: reg.ar_and(x, y, n_xz, n_yz)
        else:
            fz = lambda x, y: reg.ar_or(x, y, n_xz, n_yz)
    elif ffl[:2].lower() in ("c2", "i3"):
        if logic.lower() == "and":
            fz = lambda x, y: reg.ar_and(y, x, n_yz, n_xz)
        else:
            fz = lambda x, y: reg.ar_or(y, x, n_yz, n_xz)
    else:
        if logic.lower() == "and":
            fz = lambda x, y: reg.rr_and(x, y, n_xz, n_yz)
        else:
            fz = lambda x, y: reg.rr_or(x, y, n_xz, n_yz)

    def rhs(yz, t, x):
        y, z = yz
        dy_dt = beta * fy(kappa * x) - y
        dz_dt = gamma * (fz(x, y) - z)

        return np.array([dy_dt, dz_dt])

    return rhs


def solve_ffl(beta, gamma, kappa, n_xy, n_xz, n_yz, ffl, logic, t, t_step_down, x_0):
    """Solve an FFL. The dynamics are given by
    `rhs`, the output of `ffl_rhs()`.
    """
    if t[0] != 0:
        raise RuntimeError("time must start at zero.")

    rhs = _ffl_rhs(beta, gamma, kappa, n_xy, n_xz, n_yz, ffl, logic)

    # Integrate if we do not step down
    if t[-1] < t_step_down:
        return scipy.integrate.odeint(rhs, np.zeros(2), t, args=(x_0,))

    # Integrate up to step down
    t_during_step = np.concatenate((t[t < t_step_down], (t_step_down,)))
    yz_during_step = scipy.integrate.odeint(
        rhs, np.zeros(2), t_during_step, args=(x_0,)
    )

    # Integrate after step
    t_after_step = np.concatenate(((t_step_down,), t[t > t_step_down]))
    yz_after_step = scipy.integrate.odeint(
        rhs, yz_during_step[-1, :], t_after_step, args=(0,)
    )

    # Concatenate solutions
    if t_step_down in t:
        return np.vstack((yz_during_step[:-1, :], yz_after_step))
    else:
        return np.vstack((yz_during_step[:-1, :], yz_after_step[1:, :]))


def plot_ffl(
    beta=1.0,
    gamma=1.0,
    kappa=1.0,
    n_xy=1.0,
    n_xz=1.0,
    n_yz=1.0,
    ffl="c1",
    logic="and",
    t=np.linspace(0, 20, 200),
    t_step_down=10.0,
    x_0=1.0,
    normalized=False,
    **kwargs,
):
    yz = solve_ffl(
        beta, gamma, kappa, n_xy, n_xz, n_yz, ffl, logic, t, t_step_down, x_0
    )
    y, z = yz.transpose()

    # Generate x-values
    if t[-1] > t_step_down:
        t_x = np.array([-t_step_down / 10, 0, 0, t_step_down, t_step_down, t[-1]])
        x = np.array([0, 0, x_0, x_0, 0, 0], dtype=float)
    else:
        t_x = np.array([-t[-1] / 10, 0, 0, t[-1]])
        x = np.array([0, 0, x_0, x_0], dtype=float)

    # Add left part of y and z-values
    t = np.concatenate(((t_x[0],), t))
    y = np.concatenate(((0,), y))
    z = np.concatenate(((0,), z))

    # Normalize if necessary
    if normalized:
        x /= x.max()
        y /= y.max()
        z /= z.max()

    # Set up figure
    frame_height = kwargs.pop("frame_height", 175)
    frame_width = kwargs.pop("frame_width", 550)
    x_axis_label = kwargs.pop("x_axis_label", "dimensionless time")
    y_axis_label = kwargs.pop(
        "y_axis_label", f"{'norm. ' if normalized else ''}dimensionless conc."
    )
    x_range = kwargs.pop("x_range", [t.min(), t.max()])
    p = bokeh.plotting.figure(
        frame_height=frame_height,
        frame_width=frame_width,
        x_axis_label=x_axis_label,
        y_axis_label=y_axis_label,
        x_range=x_range,
    )

    # Column data sources
    cds = bokeh.models.ColumnDataSource(dict(t=t, y=y, z=z))
    cds_x = bokeh.models.ColumnDataSource(dict(t=t_x, x=x))

    # Populate glyphs
    colors = colorcet.b_glasbey_category10
    p.line(source=cds_x, x="t", y="x", line_width=2, color=colors[0], legend_label="x")
    p.line(source=cds, x="t", y="y", line_width=2, color=colors[1], legend_label="y")
    p.line(source=cds, x="t", y="z", line_width=2, color=colors[2], legend_label="z")

    # Allow vanishing lines by clicking legend
    p.legend.click_policy = "hide"

    return p, cds, cds_x


def _ffl_callback(
    p,
    cds,
    cds_x,
    beta,
    gamma,
    kappa,
    n_xy,
    n_xz,
    n_yz,
    ffl,
    logic,
    t_step_down,
    x_0,
    normalized,
):
    # Time points based on current axis limits
    t = np.linspace(0, p.x_range.end, 400)

    # Solve the dynamics
    yz = solve_ffl(
        beta, gamma, kappa, n_xy, n_xz, n_yz, ffl, logic, t, t_step_down, x_0
    )
    y, z = yz.transpose()

    # Generate x-values
    if t[-1] > t_step_down:
        t_x = np.array([-t_step_down / 10, 0, 0, t_step_down, t_step_down, t[-1]])
        x = np.array([0, 0, x_0, x_0, 0, 0], dtype=float)
    else:
        t_x = np.array([-t[-1] / 10, 0, 0, t[-1]])
        x = np.array([0, 0, x_0, x_0], dtype=float)

    # Add left part of y and z-values
    t = np.concatenate(((t_x[0],), t))
    y = np.concatenate(((0,), y))
    z = np.concatenate(((0,), z))

    # Normalize if necessary
    if normalized:
        x /= x.max()
        y /= y.max()
        z /= z.max()

    # Update ColumnDataSource
    cds.data = dict(t=t, y=y, z=z)
    cds_x.data = dict(t=t_x, x=x)


def _ffl_widgets():
    param_sliders_kwargs = dict(start=0.1, end=10, step=0.1, value=1, width=125)
    hill_coeff_kwags = dict(start=0.1, end=10, step=0.1, value=1, width=125)

    widgets = dict(
        beta_slider=bokeh.models.Slider(title="β", **param_sliders_kwargs),
        gamma_slider=bokeh.models.Slider(title="γ", **param_sliders_kwargs),
        kappa_slider=bokeh.models.Slider(title="κ", **param_sliders_kwargs),
        n_xy_slider=bokeh.models.Slider(title="nxy", **hill_coeff_kwags),
        n_xz_slider=bokeh.models.Slider(title="nxz", **hill_coeff_kwags),
        n_yz_slider=bokeh.models.Slider(title="nyz", **hill_coeff_kwags),
        ffl_selector=bokeh.models.Select(
            title="Circuit",
            options=[
                f"{x}-FFL" for x in ["C1", "C2", "C3", "C4", "I1", "I2", "I3", "I4"]
            ],
            value="C1-FFL",
            width=125,
        ),
        logic_selector=bokeh.models.RadioButtonGroup(
            name="Logic", labels=["AND", "OR"], active=0, width=125
        ),
        t_step_down_slider=bokeh.models.Slider(
            title="step down time", start=0.1, end=21, step=0.1, value=10, width=125
        ),
        x_0_slider=bokeh.models.Slider(
            title="x₀", start=0.1, end=10, step=0.1, value=1, width=125
        ),
        normalize_toggle=bokeh.models.Toggle(
            label="Normalize", active=False, width=125
        ),
    )

    return widgets


def ffl_app():
    """Create a Bokeh app for exploring the dynamics of feed-forward loops
    in response to a step input.

    Returns
    -------
    app : function
        The `app` function can be used to invoke a Bokeh app.

    Notes
    -----
    .. To serve the app from the command line so it has its own page
       in the browser, you can create a `.py`, say called
       `ffl_app.py`, with the following contents:

       ```
       import biocircuits.apps
       import bokeh.plotting

       app = biocircuits.apps.promiscuous_222_app()

       app(bokeh.plotting.curdoc())
       ```
       Then, from the command line, run:
       `bokeh serve --show ffl_app.py`
    .. To run the app from a Jupyter notebook, do the following in a
       code cell:
       ```
       import biocircuits.apps
       import bokeh.io

       bokeh.io.output_notebook()

       app = biocircuits.apps.ffl_app()
       bokeh.io.show(app, notebook_url='localhost:8888')

       ```
       You may need to change the `notebook_url` as necessary.
    """

    def app(doc):
        p, cds, cds_x = plot_ffl()
        widgets = _ffl_widgets()

        def _callback(attr, old, new):

            _ffl_callback(
                p,
                cds,
                cds_x,
                widgets["beta_slider"].value,
                widgets["gamma_slider"].value,
                widgets["kappa_slider"].value,
                widgets["n_xy_slider"].value,
                widgets["n_xz_slider"].value,
                widgets["n_yz_slider"].value,
                widgets["ffl_selector"].value,
                ("AND", "OR")[widgets["logic_selector"].active],
                widgets["t_step_down_slider"].value,
                widgets["x_0_slider"].value,
                widgets["normalize_toggle"].active,
            )

        for widget_name, widget in widgets.items():
            try:
                widget.on_change("value", _callback)
            except:
                widget.on_change("active", _callback)

        sliders = bokeh.layouts.row(
            bokeh.layouts.Spacer(width=30),
            bokeh.layouts.column(
                widgets["beta_slider"], widgets["gamma_slider"], widgets["kappa_slider"]
            ),
            bokeh.layouts.Spacer(width=10),
            bokeh.layouts.column(
                widgets["n_xy_slider"], widgets["n_xz_slider"], widgets["n_yz_slider"]
            ),
            bokeh.layouts.Spacer(width=10),
            bokeh.layouts.column(widgets["t_step_down_slider"], widgets["x_0_slider"]),
        )

        selectors = bokeh.layouts.column(
            widgets["ffl_selector"],
            widgets["logic_selector"],
            widgets["normalize_toggle"],
        )

        # Final layout
        layout = bokeh.layouts.column(
            p,
            bokeh.layouts.Spacer(width=20),
            bokeh.layouts.row(sliders, bokeh.layouts.Spacer(width=20), selectors),
        )

        doc.add_root(layout)

    return app



import os

import numpy as np
import pandas as pd

import eqtk

import bokeh.models
import bokeh.layouts
import bokeh.palettes
import bokeh.plotting


def _make_rxns(nA, nB, nL):
    rxns = ""
    # Dimer rxns
    for j in range(nL):
        for i in range(nA):
            rxns += f"A_{i+1} + L_{j+1} <=> D_{i+1}_{j+1}\n"

    # Trimer rxns
    for j in range(nL):
        for i in range(nA):
            for k in range(nB):
                rxns += f"D_{i+1}_{j+1} + B_{k+1} <=> T_{i+1}_{j+1}_{k+1}\n"

    return rxns


def _make_N(nA, nB, nL):
    rxns = _make_rxns(nA, nB, nL)
    N = eqtk.parse_rxns(rxns)

    # Impose ordering of names
    names = [f"A_{i+1}" for i in range(nA)]
    names += [f"B_{i+1}" for i in range(nB)]
    names += [f"L_{i+1}" for i in range(nL)]
    names += [f"D_{i+1}_{j+1}" for j in range(nL) for i in range(nA)]
    names += [
        f"T_{i+1}_{j+1}_{k+1}" for j in range(nL) for i in range(nA) for k in range(nB)
    ]

    # As a Numpy array
    return N[names].to_numpy(copy=True, dtype=float)


def _readout(epsilon, c):
    return np.dot(epsilon, c[:, -len(epsilon) :].transpose())


def _random_epsilon(nA, nB, nL):
    return np.random.dirichlet(np.ones(nA * nB * nL))


def _random_K(nA, nB, nL):
    # Dimerization equilibrium constants
    K_dimer = np.concatenate([np.random.dirichlet(np.ones(nA)) for i in range(nL)])

    # Trimerization equilibrium constants
    K_trimer = np.random.dirichlet(np.ones(nA * nB * nL))

    return np.concatenate((K_dimer, K_trimer))


def _make_c0_grid(nA, nB, nL, n, low_ligand_conc=0.001, high_ligand_conc=1000.0):
    log_low = np.log10(low_ligand_conc)
    log_high = np.log10(high_ligand_conc)
    # Ligand concentrations
    cL0 = np.logspace(log_low, log_high, n)
    cL0 = np.meshgrid(*tuple([cL0] * nL))

    # Initialize c0 and fixed_c
    c0 = np.zeros((n ** nL, nA + nB + nL + nA * nL + nA * nB * nL))
    fixed_c = np.ones_like(c0) * np.nan

    # Add ligand concentrations
    for i in range(nL):
        c0[:, i + nA + nB] = cL0[i].flatten()
        fixed_c[:, i + nA + nB] = cL0[i].flatten()

    # Random concentrations of receptors
    for i in range(nA):
        c0[:, i] = 10 ** np.random.uniform(log_low, log_high)
    for i in range(nB):
        c0[:, i + nA] = 10 ** np.random.uniform(log_low, log_high)

    return c0, fixed_c


def _make_c0_rim(n, logA, logB, log_low, log_high):
    # Initialize c0 and fixed_c
    c0 = np.zeros((2 * n, 18))
    fixed_c = np.zeros_like(c0) * np.nan

    # Ligand concentrations
    cL10 = np.concatenate(
        [[0], np.logspace(log_low, log_high, n), [10 ** log_high] * (n - 1)]
    )
    cL20 = np.concatenate(
        [[10 ** log_high] * (n - 1), np.logspace(log_high, log_low, n), [0]]
    )
    c0[:, 4] = cL10
    c0[:, 5] = cL20
    fixed_c[:, 4] = cL10
    fixed_c[:, 5] = cL20

    # Cannot fix a concentration to zero, so change fixed_c=0 values
    fixed_c[np.where(fixed_c == 0)] = np.nan

    # Concentrations of receptors
    c0[:, 0] = 10 ** logA[0]
    c0[:, 1] = 10 ** logA[1]
    c0[:, 2] = 10 ** logB[0]
    c0[:, 3] = 10 ** logB[1]

    return c0, fixed_c


def _lic_rls(s):
    a = s[0]
    b = s[-1]
    c = np.max(s)
    d = np.min(s)

    # Ensure a is the low level.
    if a > b:
        a, b = b, a

    lic = d / a - b / c
    rls = a / b

    return lic, rls


def _to_label_str(x, pretty_powers_of_ten):
    pretty_powers = {
        0.0001: "10⁻⁴",
        0.001: "10⁻³",
        0.01: "10⁻²",
        0.1: "10⁻¹",
        1.0: "10⁰",
        10.0: "10¹",
        100.0: "10²",
        1000.0: "10³",
        10000.0: "10⁴",
    }
    if pretty_powers_of_ten and 0.0001 <= x <= 10000.0:
        if x % 10 == 0 or (1 / x) % 10 == 0 or x == 1:
            return pretty_powers[x]
        else:
            pow_ten = int(np.floor(np.log10(x)))
            coeff = x / 10 ** pow_ten
            return "{0:.2f}".format(coeff) + "×" + pretty_powers[10 ** pow_ten]
    else:
        return "{0:.2e}".format(x)


def _param_dict_to_array(param_dict):
    return np.concatenate(
        (param_dict["K"], param_dict["epsilon"], param_dict["logA"], param_dict["logB"])
    )


_imbalance_params = {
    "logA": np.log10(np.array([0.070003, 0.0072265])),
    "logB": np.log10(np.array([14.505, 0.0020123])),
    "K": (
        0.55065,
        0.44935,
        0.032076,
        0.96792,
        0.10169,
        0.28462,
        0.038524,
        0.0066946,
        0.020614,
        0.27227,
        0.14488,
        0.1307,
    ),
    "epsilon": (
        0.012357,
        0.18003,
        0.48492,
        0.04415,
        0.096726,
        0.086891,
        0.047076,
        0.047853,
    ),
}

_additive_params = {
    "logA": np.log10(np.array([0.14356, 0.0031554])),
    "logB": np.log10(np.array([0.029711, 0.089175])),
    "K": (
        0.73952,
        0.26048,
        0.33582,
        0.66418,
        0.31139,
        0.10826,
        0.086969,
        0.019978,
        0.008519,
        0.092182,
        0.33719,
        0.035517,
    ),
    "epsilon": (
        0.052266,
        0.092732,
        0.35798,
        0.1635,
        0.0915,
        0.16255,
        0.076068,
        0.0034033,
    ),
}

_ratiometric_params = {
    "logA": np.log10(np.array([14.758, 85.588])),
    "logB": np.log10(np.array([107.92, 0.28909])),
    "K": (
        0.0016677,
        0.99833,
        0.71197,
        0.28803,
        0.010589,
        0.22728,
        0.24735,
        0.0040583,
        0.13076,
        0.24912,
        0.10161,
        0.029231,
    ),
    "epsilon": (
        0.098864,
        0.021308,
        0.00091882,
        0.22745,
        0.039245,
        0.16528,
        0.38461,
        0.062319,
    ),
}

_balance_params = {
    "logA": np.log10(np.array([2.2772, 270.11])),
    "logB": np.log10(np.array([0.2441, 1.2624])),
    "K": (
        0.74172,
        0.25828,
        0.66746,
        0.33254,
        0.22697,
        0.030042,
        0.22169,
        0.001393,
        0.17624,
        0.11554,
        0.047332,
        0.18079,
    ),
    "epsilon": (
        0.095076,
        0.017184,
        0.49264,
        0.032259,
        0.16782,
        0.073906,
        0.021015,
        0.1001,
    ),
}


def promiscuous_222_app(
    n=13, low_ligand_conc=0.001, high_ligand_conc=1000.0, pretty_powers_of_ten=True,
):
    """Create a Bokeh app for exploring promiscuous ligand-receptor binding
    for a system with two types of receptor A, two types of receptor B,
    and two ligands.

    Parameters
    ----------
    n : int, default 13
        Number of different ligand concentrations to consider. Ligand
        concentrations vary logarithmically over `low_ligand_conc` to
        `high_ligand_conc`.
    low_ligand_conc: float, default 0.001
        Lowest ligand concentration to consider.
    high_ligand_conc: float, default 1000.0
        Highest ligand concentration to consider.
    pretty_powers_of_ten : bool, default True
        If `True`, then attempts are made to make axis labels look
        "pretty", like `10²` instead of `1.0e+02`. Not guaranteed to
        work for all choices of ligand concentration.

    Returns
    -------
    app : function
        The `app` function can be used to invoke a Bokeh app.

    Notes
    -----
    .. To serve the app from the command line so it has its own page
       in the browser, you can create a `.py`, say called
       `promiscuous_app.py`, with the following contents:

       ```
       import biocircuits.apps
       import bokeh.plotting

       app = biocircuits.apps.promiscuous_222_app()

       app(bokeh.plotting.curdoc())
       ```
       Then, from the command line, run:
       `bokeh serve --show promiscuous_app.py`
    .. To run the app from a Jupyter notebook, do the following in a
       code cell:
       ```
       import biocircuits.apps
       import bokeh.io

       bokeh.io.output_notebook()

       app = biocircuits.apps.promiscuous_222_app()
       bokeh.io.show(app, notebook_url='localhost:8888')

       ```
       You may need to change the `notebook_url` as necessary.
    """
    log_low = np.log10(low_ligand_conc)
    log_high = np.log10(high_ligand_conc)

    def app(doc):
        param_names = [
            "K11",
            "K21",
            "K12",
            "K22",
            "K111",
            "K112",
            "K211",
            "K212",
            "K121",
            "K122",
            "K221",
            "K222",
            "eps111",
            "eps112",
            "eps211",
            "eps212",
            "eps121",
            "eps122",
            "eps221",
            "eps222",
            "log10 A1",
            "log10 A2",
            "log10 B1",
            "log10 B2",
        ]

        # Sliders
        receptor_kwargs = dict(
            start=log_low, end=log_high, step=0.1, value=0, width=100
        )
        K_dimer_kwargs = dict(start=0.01, end=1, step=0.01, value=0.25, width=100)
        K_trimer_kwargs = dict(start=0.01, end=1, step=0.01, value=0.125, width=100)
        eps_kwargs = dict(start=0, end=1, step=0.01, value=0.125, width=100)
        sliders = {}
        for name in param_names:
            if "log" in name:
                kwargs = receptor_kwargs
            elif "eps" in name:
                kwargs = eps_kwargs
            elif len(name) == 3:
                kwargs = K_dimer_kwargs
            else:
                kwargs = K_trimer_kwargs

            sliders[name] = bokeh.models.Slider(
                title=name.translate(str.maketrans("012", "₀₁₂")).replace("eps", "ε"),
                **kwargs,
            )

        # Preset parameters buttons
        random_params_button = bokeh.models.Button(
            label="random", button_type="success", width=100
        )
        additive_params_button = bokeh.models.Button(
            label="additive", button_type="primary", width=100
        )
        ratiometric_params_button = bokeh.models.Button(
            label="ratiometric", button_type="primary", width=100
        )
        imbalance_params_button = bokeh.models.Button(
            label="imbalance", button_type="primary", width=100
        )
        balance_params_button = bokeh.models.Button(
            label="balance", button_type="primary", width=100
        )

        # Generate random sequences
        n_random_selector = bokeh.models.Select(
            title="# of random param sets",
            value="10",
            options=["10", "30", "100", "300", "1000"],
            width=165,
        )
        gen_random_params_button = bokeh.models.Button(
            label="Generate random params", button_type="success", width=165
        )
        status = bokeh.models.Div(text="", width=150)

        # Save current parameters button
        save_params_button = bokeh.models.Button(
            label="save current params", button_type="warning", width=165
        )
        outfile_input = bokeh.models.TextInput(
            title="file name", value="_tmp.csv", width=165
        )

        # Load current parameters button
        load_params_button = bokeh.models.Button(
            label="load params", button_type="warning", width=165
        )
        infile_input = bokeh.models.TextInput(title="file name", value="", width=165)
        load_status = bokeh.models.Div(text="", width=165)

        # Initial concentrations
        c0, fixed_c = _make_c0_grid(2, 2, 2, n)

        # Indices for periphery
        inds = np.concatenate(
            (np.arange(n ** 2 - n, n ** 2), np.arange(n - 1, n ** 2 - n, n)[::-1])
        )

        # Stoichiometric matrix
        N = _make_N(2, 2, 2)

        def params_from_sliders():
            # Set concentrations
            c0[:, 0] = 10 ** sliders["log10 A1"].value
            c0[:, 1] = 10 ** sliders["log10 A2"].value
            c0[:, 2] = 10 ** sliders["log10 B1"].value
            c0[:, 3] = 10 ** sliders["log10 B2"].value

            # Pull values ouff sliders
            K = np.array([sliders[name].value for name in param_names[:12]])
            epsilon = np.array([sliders[name].value for name in param_names[12:20]])

            # Rescale K and epsilon as per paper
            K[:2] /= K[:2].astype(float).sum()
            K[2:4] /= K[2:4].astype(float).sum()
            K[4:] /= K[4:].astype(float).sum()
            epsilon /= epsilon.astype(float).sum()

            return c0, K, epsilon

        def solve():
            # Solve grid
            c0, K, epsilon = params_from_sliders()
            c = eqtk.fixed_value_solve(c0=c0, fixed_c=fixed_c, N=N, K=K)
            s = _readout(epsilon, c)

            # Initial concentrations for high/zero ligand
            c0_rim = np.vstack((c0[0, :], c0[0, :]))
            c0_rim[:, 4] = np.array([0, high_ligand_conc])
            c0_rim[:, 5] = np.array([high_ligand_conc, 0])
            fixed_c_rim = np.ones_like(c0_rim) * np.nan
            fixed_c_rim[0, 5] = high_ligand_conc
            fixed_c_rim[1, 4] = high_ligand_conc

            # Solve with one ligand concentration set to zero and one high
            c_extremes = eqtk.fixed_value_solve(
                c0=c0_rim, fixed_c=fixed_c_rim, N=N, K=K
            )
            s_extremes = _readout(epsilon, c_extremes)
            s_rim = np.concatenate(((s_extremes[0],), s[inds], (s_extremes[1],)))

            return s, s_rim

        # Convert x and y values to strings
        x_str = [_to_label_str(val, pretty_powers_of_ten) for val in c0[:, 4]]
        y_str = [_to_label_str(val, pretty_powers_of_ten) for val in c0[:, 5]]

        # Set the locations of the ticks
        x_range = [
            _to_label_str(val, pretty_powers_of_ten)
            for val in sorted(list(np.unique(c0[:, 4])))
        ]
        y_range = [
            _to_label_str(val, pretty_powers_of_ten)
            for val in sorted(list(np.unique(c0[:, 5])))
        ]

        # Set up figures
        p = bokeh.plotting.figure(
            x_range=x_range,
            y_range=y_range,
            frame_width=300,
            frame_height=300,
            x_axis_label="L₁",
            y_axis_label="L₂",
            tools="pan,box_zoom,wheel_zoom,reset,hover,save",
            tooltips=[("L1", "@x"), ("L2", "@y"), ("S", "@z")],
            toolbar_location="above",
            align="end",
            title="matrix",
        )

        p_rim = bokeh.plotting.figure(
            x_range=[-0.05 * 2 * n, 1.05 * 2 * n],
            y_range=[-0.25, 1.05],
            frame_width=300,
            frame_height=125,
            x_axis_label="coordinate along rim",
            y_axis_label="normalize signal",
            toolbar_location="above",
            tools="save",
            align="end",
            title="rim",
        )

        p_lic_rls = bokeh.plotting.figure(
            frame_width=300,
            frame_height=125,
            x_range=[-1.05, 1.05],
            y_range=[-0.05, 1.05],
            x_axis_label="ligand interference coefficient",
            y_axis_label="relative ligand strength",
            toolbar_location="above",
            tools="pan,box_zoom,wheel_zoom,reset,hover,tap,save",
            tooltips=[("LIC", "@lic"), ("RLS", "@rls")],
            align="end",
            title="LIC/RLS",
        )

        # Shading on rim plot
        p_rim.varea(
            [-0.05 * 2 * n, n, 1.05 * 2 * n],
            [-0.25, -0.25, -0.25],
            [0, 0, -0.25],
            fill_alpha=0.15,
            fill_color="#fc8d62",
        )
        p_rim.varea(
            [-0.05 * 2 * n, n, 1.05 * 2 * n],
            [-0.25, -0.25, -0.25],
            [-0.25, 0, 0],
            fill_alpha=0.15,
            fill_color="#8da0cb",
        )
        p_rim.line([-0.05 * 2 * n, 1.05 * 2 * n], [0, 0], color="black")
        L1_label = bokeh.models.Label(
            x=1.02 * 2 * n,
            y=-0.175,
            text="L₁",
            text_color="#8da0cb",
            text_align="right",
            text_font_size="8pt",
        )
        L2_label = bokeh.models.Label(
            x=-0.02 * n, y=-0.175, text="L₂", text_color="#fc8d62", text_font_size="8pt"
        )
        p_rim.add_layout(L1_label)
        p_rim.add_layout(L2_label)

        # Set up the data sources
        source = bokeh.models.ColumnDataSource(
            dict(x_str=[], y_str=[], z_val=[], x=[], y=[], z=[])
        )
        source_rim = bokeh.models.ColumnDataSource(dict(x=[], y=[]))
        source_rim_area_L1 = bokeh.models.ColumnDataSource(dict(x=[], y=[]))
        source_rim_area_L2 = bokeh.models.ColumnDataSource(dict(x=[], y=[]))

        source_lic_rls_curr = bokeh.models.ColumnDataSource(dict(lic=[], rls=[]))

        source_lic_rls_data = {name: [] for name in param_names}
        source_lic_rls_data["lic"] = []
        source_lic_rls_data["rls"] = []
        source_lic_rls_cum = bokeh.models.ColumnDataSource(source_lic_rls_data)

        # Color mapper
        mapper = bokeh.models.LinearColorMapper(
            palette=bokeh.palettes.Viridis256, low=0, high=1
        )

        # Put in the glyphs, coloring by z_val
        p.rect(
            x="x_str",
            y="y_str",
            width=1,
            height=1,
            source=source,
            fill_color={"field": "z_val", "transform": mapper},
            line_color=None,
        )

        # Glyphs for periphery
        p_rim.circle(x="x", y="y", source=source_rim)

        # LIC/RLS glyphs
        p_lic_rls.circle(
            x="lic", y="rls", source=source_lic_rls_cum, size=3, fill_alpha=0,
        )
        p_lic_rls.circle(
            x="lic",
            y="rls",
            source=source_lic_rls_curr,
            size=3,
            fill_alpha=0,
            color="tomato",
        )

        # Style
        p.grid.grid_line_color = None
        p.axis.axis_line_color = None
        p.axis.major_tick_line_color = None
        p.axis.major_label_text_font_size = "9px"
        p.axis.major_label_standoff = 0
        if n != 7:
            p.xaxis.major_label_orientation = np.pi / 3

        # Callbacks
        def update(attr, old, new):
            s, s_rim = solve()

            # Scale to get full range of colormapper
            z_val = (s - s.min()) / (s.max() - s.min())

            # Compute LIC and RLS
            lic, rls = _lic_rls(s_rim)

            # All measurements
            s_all = np.concatenate((s, s_rim))

            # Normalize s_rim
            s_rim = (s_rim - s_all.min()) / (s_all.max() - s_all.min())

            # Update sources
            source.data = dict(
                x_str=x_str, y_str=y_str, z_val=z_val, x=c0[:, 4], y=c0[:, 5], z=s,
            )
            source_rim.data = dict(x=np.arange(2 * n + 1), y=s_rim)

            data_dict = {name: [sliders[name].value] for name in param_names}
            data_dict["lic"] = [lic]
            data_dict["rls"] = [rls]
            source_lic_rls_cum.stream(data_dict)
            source_lic_rls_curr.data = dict(lic=[lic], rls=[rls])

        def random_parameters(event=None):
            K_dict = {name: [] for name in param_names[:12]}
            eps_dict = {name: [] for name in param_names[12:20]}
            receptor_dict = {
                "log10 A1": [],
                "log10 A2": [],
                "log10 B1": [],
                "log10 B2": [],
            }
            lic_rls_dict = {"lic": [], "rls": []}

            for i in range(int(n_random_selector.value)):
                K = _random_K(2, 2, 2)
                epsilon = _random_epsilon(2, 2, 2)
                logA = np.random.uniform(log_low, log_high, size=2)
                logB = np.random.uniform(log_low, log_high, size=2)
                c0_rim, fixed_c_rim = _make_c0_rim(n, logA, logB, log_low, log_high)

                # Solve along the rim
                c_rim = eqtk.fixed_value_solve(c0=c0_rim, fixed_c=fixed_c_rim, N=N, K=K)
                s_rim = _readout(epsilon, c_rim)

                # Compute LIC and RLS
                lic, rls = _lic_rls(s_rim)

                for i, name in enumerate(param_names[:12]):
                    K_dict[name].append(K[i])
                for i, name in enumerate(param_names[12:20]):
                    eps_dict[name].append(epsilon[i])
                receptor_dict["log10 A1"].append(logA[0])
                receptor_dict["log10 A2"].append(logA[1])
                receptor_dict["log10 B1"].append(logB[0])
                receptor_dict["log10 B2"].append(logB[1])
                lic_rls_dict["lic"].append(lic)
                lic_rls_dict["rls"].append(rls)

            data_dict = {**lic_rls_dict, **K_dict, **eps_dict, **receptor_dict}
            source_lic_rls_cum.stream(data_dict)

        def save_current_parameters(event=None):
            c0, K, epsilon = params_from_sliders()
            receptor_dict = {
                "A1": [c0[0, 0]],
                "A2": [c0[0, 1]],
                "B1": [c0[0, 2]],
                "B2": [c0[0, 3]],
            }
            K_dict = {name: [K[i]] for i, name in enumerate(param_names[:12])}
            eps_dict = {name: [epsilon[i]] for i, name in enumerate(param_names[12:20])}

            df = pd.DataFrame({**receptor_dict, **K_dict, **eps_dict})
            fname = outfile_input.value
            if os.path.isfile(fname):
                out_str = df.to_csv(index=False, header=False)
                with open(fname, "a") as f:
                    f.write(out_str)
            else:
                df.to_csv(fname, index=False)

        def load_parameters(event=None):
            load_status.text = "<center>Loading....</center>"
            try:
                df = pd.read_csv(infile_input.value)
                df["log10 A1"] = np.log10(df["A1"])
                df["log10 A2"] = np.log10(df["A2"])
                df["log10 B1"] = np.log10(df["B1"])
                df["log10 B2"] = np.log10(df["B2"])

                K_dict = {name: [] for name in param_names[:12]}
                eps_dict = {name: [] for name in param_names[12:20]}
                receptor_dict = {
                    "log10 A1": [],
                    "log10 A2": [],
                    "log10 B1": [],
                    "log10 B2": [],
                }
                lic_rls_dict = {"lic": [], "rls": []}

                for i, r in df.iterrows():
                    K = np.array([r[name] for name in param_names[:12]])
                    epsilon = np.array([r[name] for name in param_names[12:20]])
                    logA = np.array([r["log10 A1"], r["log10 A2"]])
                    logB = np.array([r["log10 B1"], r["log10 B2"]])

                    c0_rim, fixed_c_rim = _make_c0_rim(n, logA, logB, log_low, log_high)

                    # Solve along the rim
                    c_rim = eqtk.fixed_value_solve(
                        c0=c0_rim, fixed_c=fixed_c_rim, N=N, K=K
                    )
                    s_rim = _readout(epsilon, c_rim)

                    # Compute LIC and RLS
                    lic, rls = _lic_rls(s_rim)

                    for i, name in enumerate(param_names[:12]):
                        K_dict[name].append(K[i])
                    for i, name in enumerate(param_names[12:20]):
                        eps_dict[name].append(epsilon[i])
                    receptor_dict["log10 A1"].append(logA[0])
                    receptor_dict["log10 A2"].append(logA[1])
                    receptor_dict["log10 B1"].append(logB[0])
                    receptor_dict["log10 B2"].append(logB[1])
                    lic_rls_dict["lic"].append(lic)
                    lic_rls_dict["rls"].append(rls)

                data_dict = {**lic_rls_dict, **K_dict, **eps_dict, **receptor_dict}
                source_lic_rls_cum.stream(data_dict)

                load_status.text = "<center>Load successful.</center>"
            except:
                load_status.text = "<center>Load failed.</center>"

        def remove_on_changes():
            for param in param_names:
                sliders[param].remove_on_change("value_throttled", update)

        def add_on_changes():
            for param in param_names:
                sliders[param].on_change("value_throttled", update)

        def params_reset_random(event=None):
            K = _random_K(2, 2, 2)
            epsilon = _random_epsilon(2, 2, 2)
            logA = np.random.uniform(log_low, log_high, size=2)
            logB = np.random.uniform(log_low, log_high, size=2)

            reset_params(np.concatenate((K, epsilon, logA, logB)))

        def params_reset_additive(event=None):
            reset_params(_param_dict_to_array(_additive_params))

        def params_reset_ratiometric(event=None):
            reset_params(_param_dict_to_array(_ratiometric_params))

        def params_reset_imbalance(event=None):
            reset_params(_param_dict_to_array(_imbalance_params))

        def params_reset_balance(event=None):
            reset_params(_param_dict_to_array(_balance_params))

        def reset_params(params):
            for name, val in zip(param_names, params):
                sliders[name].value = val
            update(None, None, None)

        def params_reset_from_selection(attr, old, new):
            try:
                ind = new[0]
                for name in param_names:
                    sliders[name].value = source_lic_rls_cum.data[name][ind]

                update(None, None, None)
            except:
                pass

        # Link sliders and buttons
        add_on_changes()
        random_params_button.on_click(params_reset_random)
        additive_params_button.on_click(params_reset_additive)
        ratiometric_params_button.on_click(params_reset_ratiometric)
        imbalance_params_button.on_click(params_reset_imbalance)
        balance_params_button.on_click(params_reset_balance)
        gen_random_params_button.on_click(random_parameters)
        save_params_button.on_click(save_current_parameters)
        load_params_button.on_click(load_parameters)

        # Run update to populate plots
        update(None, None, None)

        source_lic_rls_cum.selected.on_change("indices", params_reset_from_selection)

        # Build layout
        K_dimer = [sliders[name] for name in param_names[:4]]
        receptors = [sliders[name] for name in param_names if "log10" in name]
        col1 = K_dimer + receptors
        col2 = [sliders[name] for name in param_names[4:12]]
        col3 = [sliders[name] for name in param_names[12:20]]

        slider_row = bokeh.layouts.row(
            bokeh.layouts.column(*col1),
            bokeh.layouts.Spacer(width=10),
            bokeh.layouts.column(*col2),
            bokeh.layouts.Spacer(width=10),
            bokeh.layouts.column(*col3),
        )

        row1 = bokeh.layouts.row(p, bokeh.layouts.Spacer(width=25), slider_row)

        buttons = bokeh.layouts.column(
            bokeh.models.Spacer(height=20),
            bokeh.models.Div(text="<b>Preset parameters</b>", width=200),
            bokeh.layouts.row(
                additive_params_button, ratiometric_params_button, random_params_button,
            ),
            bokeh.layouts.row(imbalance_params_button, balance_params_button,),
        )

        row2 = bokeh.layouts.row(
            bokeh.layouts.Spacer(width=15),
            p_rim,
            bokeh.layouts.Spacer(width=25),
            buttons,
        )

        random_params = bokeh.layouts.row(
            bokeh.layouts.Spacer(height=20),
            bokeh.layouts.column(
                bokeh.layouts.Spacer(height=18), gen_random_params_button,
            ),
            bokeh.layouts.Spacer(width=15),
            n_random_selector,
        )

        save_params = bokeh.layouts.row(
            bokeh.layouts.column(bokeh.layouts.Spacer(height=18), save_params_button),
            bokeh.layouts.Spacer(width=15),
            outfile_input,
        )

        load_params = bokeh.layouts.row(
            bokeh.layouts.column(
                bokeh.layouts.Spacer(height=18), load_params_button, load_status
            ),
            bokeh.layouts.Spacer(width=15),
            infile_input,
        )

        params_col = bokeh.layouts.column(
            bokeh.models.Spacer(height=15),
            random_params,
            bokeh.layouts.Spacer(height=15),
            save_params,
            bokeh.layouts.Spacer(height=15),
            load_params,
        )

        row3 = bokeh.layouts.row(
            bokeh.layouts.Spacer(width=15),
            p_lic_rls,
            bokeh.layouts.Spacer(width=25),
            params_col,
        )

        instructions = bokeh.models.Div(
            text="""
            <b><center>Instructions</center></b>
            <div style="border: 1px solid black;">
            <ul>
            <li>In generating the plots, the values of the parameters of the sliders are rescaled such that K₁₁ + K₁₂ = 1, K₂₁ + K₂₂ = 1, the sum of all K's with three indices equals one, and the sum of all ε's equals 1.</li>
            <br />
            <li>The matrix plot uses color coding to show the magnitude of the response for various ligand concentrations. All units are arbitrary. The lowest signal is colored purple, and the highest is colored yellow.</li>
            <br />
            <li>The rim plot shows the magnitude of the response going from the upper left to upper right to lower right of the matrix plot, shouldered on the beginning and end by L₁ = 0 and L₂ = 0, respectively. the magnitude of the response is normalized such that the minimum value on the matrix plots is zero and the maximum value on the matrix plot is one.</li>
            <br />
            <li>The LIC/RLS plot shows the relative ligand strength (RLS) plotted against the ligand interference coefficient (LIC) for every parameter set encountered since the dashboard was launched. The red point on the LIC/RLS plot corresponds to the current parameter values of the sliders. All previously encountered parameter values are shown in blue. Click on a blue point to set the parameter values corresponding to that point.</li>
            <br />
            <li>Adjust sliders to change parameter values. The response to the sliders is throttled, so plots will only respond when you release our mouse button.</li>
            <br />
            <li>You can load preset parameter values by clicking the blue buttons. Clicking the green "random" button will adjust the sliders and plots for a random set of parameters.</li>
            <br />
            <li>To expedite exploration of parameters, you can generate many random parameter sets and populate the LIC/RLS plot with the results by clicking the "generate random params" button. The values of the sliders and the matrix and rim plot will not change, but you can click on any of the blue dots added to the LIC/RLS plot to set the sliders and plots to those parameters. You can choose how many sets of parameters you which to generate from the "# of random param sets" pull-down menu. You may have to wait several seconds if you generate more than 30 sets at a time.</li>
            <br />
            <li>To store the current parameter values, click the "save current params" button. The parameter values will be saved to a CSV file given in the adjacent "file name" text box. If the file already exists, the current parameter values will be appended.</li>
            <br />
            <li>To load sets of parameters, click the "load params" button and the parameters in file specified in the adjacent "file name" text box will be loaded. The parameters are loaded and a corresponding glyph is added to the LIC/RLS plot. The format of the input file is the same as the output when you click "save current params."</li>
            </ul>
            </div>
        """,
            width=560,
        )

        row4 = bokeh.layouts.row(bokeh.layouts.Spacer(width=110), instructions)

        doc.add_root(
            bokeh.layouts.column(
                row1,
                bokeh.layouts.Spacer(height=30),
                row2,
                bokeh.layouts.Spacer(height=30),
                row3,
                bokeh.layouts.Spacer(height=30),
                row4,
            )
        )

    return app



import numpy as np
import scipy.integrate


def _dc_dt(
    c,
    t,
    x,
    derivs_0,
    derivs_L,
    diff_coeff_fun,
    diff_coeff_params,
    rxn_fun,
    rxn_params,
    n_species,
    h,
):
    """
    Time derivative of concentrations in an R-D system
    for constant flux BCs.

    Parameters
    ----------
    c : ndarray, shape (n_species * n_gridpoints)
        The concentration of the chemical species interleaved in a
        a NumPy array.  The interleaving allows us to take advantage
        of the banded structure of the Jacobian when using the
        Hindmarsh algorithm for integrating in time.
    t : float
        Time.
    derivs_0 : ndarray, shape (n_species)
        derivs_0[i] is the value of the diffusive flux,
        D dc_i/dx, at x = 0, the leftmost boundary of the domain of x.
    derivs_L : ndarray, shape (n_species)
        derivs_0[i] is the value of the diffusive flux,
        D dc_i/dx, at x = L, the rightmost boundary of the domain of x.
    diff_coeff_fun : function
        Function of the form diff_coeff_fun(c_tuple, t, x, *diff_coeff_params).
        Returns an tuple where entry i is a NumPy array containing
        the diffusion coefficient of species i at the grid points.
        c_tuple[i] is a NumPy array containing the concentrations of
        species i at the grid poitns.
    diff_coeff_params : arbitrary
        Tuple of parameters to be passed into diff_coeff_fun.
    rxn_fun : function
        Function of the form rxn_fun(c_tuple, t, *rxn_params).
        Returns an tuple where entry i is a NumPy array containing
        the net rate of production of species i by chemical reaction
        at the grid points.  c_tuple[i] is a NumPy array containing
        the concentrations of species i at the grid poitns.
    rxn_params : arbitrary
        Tuple of parameters to be passed into rxn_fun.
    n_species : int
        Number of chemical species.
    h : float
        Grid spacing (assumed to be constant)

    Returns
    -------
    dc_dt : ndarray, shape (n_species * n_gridpoints)
        The time derivatives of the concentrations of the chemical
        species at the grid points interleaved in a NumPy array.
    """
    # Tuple of concentrations
    c_tuple = tuple([c[i::n_species] for i in range(n_species)])

    # Compute diffusion coefficients
    D_tuple = diff_coeff_fun(c_tuple, t, x, *diff_coeff_params)

    # Compute reaction terms
    rxn_tuple = rxn_fun(c_tuple, t, *rxn_params)

    # Return array
    conc_deriv = np.empty_like(c)

    # Convenient array for storing concentrations
    da_dt = np.empty(len(c_tuple[0]))

    # Useful to have square of grid spacing around
    h2 = h ** 2

    # Compute diffusion terms (central differencing w/ Neumann BCs)
    for i in range(n_species):
        # View of concentrations and diffusion coeff. for convenience
        a = np.copy(c_tuple[i])
        D = np.copy(D_tuple[i])

        # Time derivative at left boundary
        da_dt[0] = D[0] / h2 * 2 * (a[1] - a[0] - h * derivs_0[i])

        # First derivatives of D and a
        dD_dx = (D[2:] - D[:-2]) / (2 * h)
        da_dx = (a[2:] - a[:-2]) / (2 * h)

        # Time derivative for middle grid points
        da_dt[1:-1] = D[1:-1] * np.diff(a, 2) / h2 + dD_dx * da_dx

        # Time derivative at left boundary
        da_dt[-1] = D[-1] / h2 * 2 * (a[-2] - a[-1] + h * derivs_L[i])

        # Store in output array with reaction terms
        conc_deriv[i::n_species] = da_dt + rxn_tuple[i]

    return conc_deriv


def rd_solve(
    c_0_tuple,
    t,
    L=1,
    derivs_0=0,
    derivs_L=0,
    diff_coeff_fun=None,
    diff_coeff_params=(),
    rxn_fun=None,
    rxn_params=(),
    rtol=1.49012e-8,
    atol=1.49012e-8,
):
    """Solve a system of reaction-diffusion equations in space and time.

    Parameters
    ----------
    c_0_tuple : tuple
        c_0_tuple[i] is a NumPy array of length n_gridpoints with the
        initial concentrations of chemical species i at the grid points.
    t : ndarray
        An array of time points for which the solution is desired.
    L : float
        Total length of the x-domain.
    derivs_0 : ndarray, shape (n_species)
        derivs_0[i] is the value of dc_i/dx at x = 0.
    derivs_L : ndarray, shape (n_species)
        derivs_L[i] is the value of dc_i/dx at x = L, the rightmost
        boundary of the domain of x.
    diff_coeff_fun : function
        Function of the form
        diff_coeff_fun(c_tuple, t, x, *diff_coeff_params).
        Returns an tuple where entry i is a NumPy array containing
        the diffusion coefficient of species i at the grid points.
        c_tuple[i] is a NumPy array containing the concentrations of
        species i at the grid points.
    diff_coeff_params : arbitrary
        Tuple of parameters to be passed into diff_coeff_fun.
    rxn_fun : function
        Function of the form rxn_fun(c_tuple, t, *rxn_params).
        Returns an tuple where entry i is a NumPy array containing
        the net rate of production of species i by chemical reaction
        at the grid points.  c_tuple[i] is a NumPy array containing
        the concentrations of species i at the grid poitns.
    rxn_params : arbitrary
        Tuple of parameters to be passed into rxn_fun.
    rtol : float
        Relative tolerance for solver.  Default os odeint's default.
    atol : float
        Absolute tolerance for solver.  Default os odeint's default.

    Returns
    -------
    c_tuple : tuple
        c_tuple[i] is a NumPy array of shape (len(t), n_gridpoints)
        with the initial concentrations of chemical species i at
        the grid points over time.

    Notes
    -----
    .. When intergrating for long times near a steady state, you
       may need to lower the absolute tolerance (atol) because the
       solution does not change much over time and it may be difficult
       for the solver to maintain tight tolerances.
    """
    # Number of grid points
    n_gridpoints = len(c_0_tuple[0])

    # Number of chemical species
    n_species = len(c_0_tuple)

    # Grid spacing
    h = L / (n_gridpoints - 1)

    # Grid points
    x = np.linspace(0, L, n_gridpoints)

    # Set up boundary conditions
    if np.isscalar(derivs_0):
        derivs_0 = np.array(n_species * [derivs_0])
    if np.isscalar(derivs_L):
        derivs_L = np.array(n_species * [derivs_L])

    # Set up parameters to be passed in to _dc_dt
    params = (
        x,
        derivs_0,
        derivs_L,
        diff_coeff_fun,
        diff_coeff_params,
        rxn_fun,
        rxn_params,
        n_species,
        h,
    )

    # Set up initial condition
    c0 = np.empty(n_species * n_gridpoints)
    for i in range(n_species):
        c0[i::n_species] = c_0_tuple[i]

    # Solve using odeint, taking advantage of banded structure
    c = scipy.integrate.odeint(
        _dc_dt, c0, t, args=params, ml=n_species, mu=n_species, rtol=rtol, atol=atol
    )

    return tuple([c[:, i::n_species] for i in range(n_species)])


def constant_diff_coeffs(c_tuple, t, x, diff_coeffs):
    """Function for use with `rd_solve()` for constant diffusion
    coefficients.

    Parameters
    ----------
    c_tuple : tuple
        c_tuple[i] is a NumPy array containing the concentrations of
        species i at the grid points.
    t : dummy argument
        Dummy argument for the time.
    x : dummy argument
        Dummy argument for position.
    diff_coeffs : tuple
        diff_coeffs[i] is the diffusion coefficient of species i.

    Returns
    -------
    output : tuple or NumPy arrays
        A tuple containing the (constant) diffusion coeffient for all
        species at all positions.

    Notes
    -----
    .. The `t` and `x` arguments are dummy arguments so that the call
       signature matches what is required by `rd_solve()`.

    """
    n = len(c_tuple[0])
    return tuple([diff_coeffs[i] * np.ones(n) for i in range(len(c_tuple))])



import numpy as np
import scipy.integrate
import scipy.interpolate


def ddeint(func, y0, t, tau, args=(), y0_args=(), n_time_points_per_step=None):
    """Integrate a system of delay differential equations defined by
        y' = f(t, y, y(t-tau1), y(t-tau2), ...)
    using the method of steps. All tau's are assumed constant.

    Parameters
    ----------
    func : function, call signature func(y, t, y_past, *args)
        Function specifying the right hand side of the system of DDEs.
        Assuming we have a system of `n` DDEs, its arguments are:
        y : Numpy array, shape (n, )
            The current value of y.
        t : float
            The current time.
        y_past : function, call signature y_past(t)
            Function used to compute values of y in the past. This is
            not specified by the user, but called as, e.g.,
            `y_past(t-tau)` within `func`. The function is automatically
            generated using interpolants.
        args : tuple
            Tuple of any other arguments to be passed `func`. Note that
            the values of the delays, `tau`, are usually included in
            `args`.
    y0 : function, call signature y0(t, *y0_args)
        A function to compute the pre- time = t[0] values of `y`.
    t : array_like
        The time points for which the solution of the DDEs is to be
        returned.
    tau : float or array_like
        Set of time delays for DDEs. Only the shortest and longest are
        used.
    args : tuple, default ()
        Tuple of arguments to be passed to `func`.
    y0_args : tuple, default ()
        Tuple of arguments to be passed to `y0`.
    n_time_points_per_step : int, default 200
        The number of time points to store the solution for each step.
        These points are then used to compute an interpolant.

    Returns
    -------
    y : array, shape (len(t), len(y0))
    Array containing the value of y for each desired time in t.

    Notes
    -----
    .. Uses `scipy.integrate.odeint()` to integrate each step. To
       compute the values of `y` at time points from a previous step,
       uses a cubic B-spline interpolant of the solution from the
       previous step.
    .. `n_time_points_per_step` may be adjusted downward if the value
       of `y` does not change rapidly for a given step, but should be
       large enough to be able to capture the dynamics.

    """
    if np.isscalar(tau):
        tau = np.array([tau])
    else:
        tau = np.array(tau)

    if (tau <= 0).any():
        raise RuntimeError("All tau's must be greater than zero.")

    tau_short = np.min(tau)
    tau_long = np.max(tau)

    if n_time_points_per_step is None:
        n_time_points_per_step = max(
            int(1 + len(t) / (t.max() - t.min()) * tau_long), 20
        )

    t0 = t[0]

    # Past function for the first step
    y_past = lambda time_point: y0(time_point, *y0_args)

    # Integrate first step
    t_step = np.linspace(t0, t0 + tau_short, n_time_points_per_step)
    y = scipy.integrate.odeint(func, y_past(t0), t_step, args=(y_past,) + args)

    # Store result from integration
    y_dense = y.copy()
    t_dense = t_step.copy()

    # Get dimension of problem for convenience
    n = y.shape[1]

    # Integrate subsequent steps
    j = 1
    while t_step[-1] < t[-1]:
        t_start = max(t0, t_step[-1] - tau_long)
        i = np.searchsorted(t_dense, t_start, side="left")
        t_interp = t_dense[i:]
        y_interp = y_dense[i:, :]

        # Make B-spline
        tck = [scipy.interpolate.splrep(t_interp, y_interp[:, i]) for i in range(n)]

        # Interpolant of y from previous step
        y_past = (
            lambda time_point: np.array(
                [scipy.interpolate.splev(time_point, tck[i]) for i in range(n)]
            )
            if time_point > t0
            else y0(time_point, *y0_args)
        )

        # Integrate this step
        t_step = np.linspace(
            t0 + j * tau_short, t0 + (j + 1) * tau_short, n_time_points_per_step
        )
        y = scipy.integrate.odeint(func, y[-1, :], t_step, args=(y_past,) + args)

        # Store the result
        y_dense = np.append(y_dense, y[1:, :], axis=0)
        t_dense = np.append(t_dense, t_step[1:])

        j += 1

    # Interpolate solution for returning
    y_return = np.empty((len(t), n))
    for i in range(n):
        tck = scipy.interpolate.splrep(t_dense, y_dense[:, i])
        y_return[:, i] = scipy.interpolate.splev(t, tck)

    return y_return



try:
    import multiprocess
except:
    import multiprocessing as multiprocess

import warnings

import numpy as np
import numba

try:
    import tqdm
    import tqdm.notebook

    has_tqdm = True
except:
    has_tqdm = False


@numba.njit
def _sample_discrete(probs, probs_sum):
    q = np.random.rand() * probs_sum

    i = 0
    p_sum = 0.0
    while p_sum < q:
        p_sum += probs[i]
        i += 1
    return i - 1


@numba.njit
def _sum(ar):
    return ar.sum()


@numba.njit
def _draw_time(props_sum):
    return np.random.exponential(1 / props_sum)


def _gillespie_draw(propensity_func, propensities, population, t, args):
    """
    Draws a reaction and the time it took to do that reaction.
    """
    # Update propensities
    propensity_func(propensities, population, t, *args)

    # Sum of propensities
    props_sum = _sum(propensities)

    # Compute time
    time = _draw_time(props_sum)

    # Draw reaction given propensities
    rxn = _sample_discrete(propensities, props_sum)

    return rxn, time


def _gillespie_trajectory(
    propensity_func, update, population_0, time_points, draw_fun, args=()
):
    # Number of species
    n_species = update.shape[1]

    @numba.njit
    def _copy_population(population_previous, population):
        for i in range(n_species):
            population_previous[i] = population[i]

    # Initialize output
    pop_out = np.empty((len(time_points), update.shape[1]), dtype=np.int64)

    # Initialize and perform simulation
    j_time = 1
    j = 0
    t = time_points[0]
    population = population_0.copy()
    population_previous = population_0.copy()
    pop_out[0, :] = population
    propensities = np.zeros(update.shape[0])
    while j < len(time_points):
        while t < time_points[j_time]:
            # draw the event and time step
            event, dt = draw_fun(propensity_func, propensities, population, t, args)

            # Update the population
            _copy_population(population_previous, population)
            population += update[event, :]

            # Increment time
            t += dt

        # Update the index (Have to be careful about types for Numba)
        j = np.searchsorted((time_points > t).astype(np.int64), 1)

        # Update the population
        for k in np.arange(j_time, min(j, len(time_points))):
            pop_out[k, :] = population_previous

        # Increment index
        j_time = j

    return pop_out, None


@numba.njit
def _gillespie_trajectory_report_time_points(
    propensity_func, update, population_0, time_points, draw_fun, args=()
):
    # Number of iterations before concatenating arrays
    n_iter = 1000

    # Initialize arrays for storing trajectories
    pop = np.empty((n_iter, update.shape[1]), dtype=np.int64)
    tp = np.empty(n_iter, dtype=np.float64)

    # Initialize and perform simulation
    t = time_points[0]
    population = population_0.copy()
    time_points_out = np.array([t])
    pop_out = population_0.copy().reshape((1, len(population)))
    propensities = np.zeros(update.shape[0])
    j = 0
    while t < time_points[-1]:
        i = 0
        while i < n_iter and t < time_points[-1]:
            # draw the event and time step
            event, dt = draw_fun(propensity_func, propensities, population, t, args)

            # Update the population
            population += update[event, :]
            pop[i, :] = population

            # Increment time
            t += dt
            tp[i] = t

            # Increment indexes
            i += 1
            j += 1

        # Add this subtrajectory to output
        pop_out = np.concatenate((pop_out, pop))
        time_points_out = np.concatenate((time_points_out, tp))

        # Reset index
        i = 0

    return pop_out[:j, :], time_points_out[:j]


def _gillespie_ssa(
    propensity_func,
    update,
    population_0,
    time_points,
    return_time_points=False,
    size=1,
    args=(),
    progress_bar=False,
):
    """
    Uses the Gillespie stochastic simulation algorithm to sample
    from probability distribution of particle counts over time.

    Parameters
    ----------
    propensity_func : function
        Function with call signature
        `propensity_func(propensities, population, t, *args) that takes
        the current propensities and population of particle counts and
        updates the propensities for each reaction. It does not return
        anything.
    update : ndarray, shape (num_reactions, num_chemical_species)
        Entry i, j gives the change in particle counts of species j
        for chemical reaction i.
    population_0 : array_like, shape (num_chemical_species)
        Array of initial populations of all chemical species.
    time_points : array_like, shape (num_time_points,)
        Array of points in time for which to sample the probability
        distribution.
    return_time_points : bool, default False
        If True, returns a trajectory and the time points of the
        trajectory, going from time_points[0] to time_points[-1].
    size : int, default 1
        Number of trajectories to sample.
    args : tuple, default ()
        The set of parameters to be passed to propensity_func.
    progress_bar : str or bool, default False
        If True, use standard tqdm progress bar. If 'notebook', use
        tqdm.notebook progress bar. If False, no progress bar.

    Returns
    -------
    sample : ndarray, shape (size, num_time_points, num_chemical_species)
        Entry i, j, k is the count of chemical species k at time
        time_points[j] for trajectory i.
    """
    # Get number of species
    n_species = update.shape[1]

    # Make sure input population has correct dimensions
    if n_species != len(population_0):
        raise RuntimeError(
            "Number of rows in `update` must equal length of `population_0."
        )

    @numba.njit
    def _copy_population(population_previous, population):
        for i in range(n_species):
            population_previous[i] = population[i]

    # Build trajectory function based on if propensity function is jitted
    if type(propensity_func) == numba.core.registry.CPUDispatcher:

        @numba.njit
        def _draw(propensities, population, t):
            """
            Draws a reaction and the time it took to do that reaction.
            """
            # Compute propensities
            propensity_func(propensities, population, t, *args)

            # Sum of propensities
            props_sum = np.sum(propensities)

            # Compute time
            time = np.random.exponential(1 / props_sum)

            # Draw reaction given propensities
            rxn = _sample_discrete(propensities, props_sum)

            return rxn, time

        if return_time_points:

            @numba.njit
            def _traj():
                # Number of iterations before concatenating arrays
                n_iter = 1000

                # Initialize arrays for storing trajectories
                pop = np.empty((n_iter, update.shape[1]), dtype=np.int64)
                tp = np.empty(n_iter, dtype=np.float64)

                # Initialize and perform simulation
                t = time_points[0]
                population = population_0.copy()
                time_points_out = np.array([t])
                pop_out = population_0.copy().reshape((1, len(population)))
                propensities = np.zeros(update.shape[0])
                j = 0
                while t < time_points[-1]:
                    i = 0
                    while i < n_iter and t < time_points[-1]:
                        # draw the event and time step
                        event, dt = _draw(propensities, population, t)

                        # Update the population
                        population += update[event, :]
                        pop[i, :] = population

                        # Increment time
                        t += dt
                        tp[i] = t

                        # Increment indexes
                        i += 1
                        j += 1

                    # Add this subtrajectory to output
                    pop_out = np.concatenate((pop_out, pop))
                    time_points_out = np.concatenate((time_points_out, tp))

                    # Reset index
                    i = 0

                return pop_out[:j, :], time_points_out[:j]

        else:

            @numba.njit
            def _traj():
                # Initialize output
                pop_out = np.empty((len(time_points), update.shape[1]), dtype=np.int64)

                # Initialize and perform simulation
                j_time = 1
                j = 0
                t = time_points[0]
                population = population_0.copy()
                population_previous = population_0.copy()
                pop_out[0, :] = population
                propensities = np.zeros(update.shape[0])
                while j < len(time_points):
                    while t < time_points[j_time]:
                        # draw the event and time step
                        event, dt = _draw(propensities, population, t)

                        # Update the population
                        _copy_population(population_previous, population)
                        population += update[event, :]

                        # Increment time
                        t += dt

                    # Update the index (Be careful about types for Numba)
                    j = np.searchsorted((time_points > t).astype(np.int64), 1)

                    # Update the population
                    for k in np.arange(j_time, min(j, len(time_points))):
                        pop_out[k, :] = population_previous

                    # Increment index
                    j_time = j

                return pop_out, None

    else:
        if return_time_points:

            def traj():
                return _gillespie_trajectory_report_time_points(
                    propensity_func,
                    update,
                    population_0,
                    time_points,
                    _gillespie_draw,
                    args=args,
                )

        else:

            def _traj():
                return _gillespie_trajectory(
                    propensity_func,
                    update,
                    population_0,
                    time_points,
                    _gillespie_draw,
                    args=args,
                )

    # Initialize output
    pop_out = [
        np.empty((len(time_points), update.shape[1]), dtype=np.int64)
        for _ in range(size)
    ]
    t_out = [None for _ in range(size)]

    # Show progress bar
    iterator = range(size)
    if progress_bar == "notebook":
        if has_tqdm:
            iterator = tqdm.notebook.tqdm(range(size))
        else:
            warning.warn("tqdm not installed or up-to-date; skipping progress bar.")
    elif progress_bar:
        if has_tqdm:
            iterator = tqdm.tqdm(range(size))
        else:
            warning.warn("tqdm not installed or up-to-date; skipping progress bar.")

    # Perform the simulations
    for i in iterator:
        pop_out[i], t_out[i] = _traj()

    return pop_out, t_out


def _gillespie_multi_fn(args):
    """Convenient function for multithreading."""
    return _gillespie_ssa(*args)


def gillespie_ssa(
    propensity_func,
    update,
    population_0,
    time_points,
    return_time_points=False,
    size=1,
    args=(),
    n_threads=1,
    progress_bar=False,
):
    """
    Uses the Gillespie stochastic simulation algorithm to sample
    from probability distribution of particle counts over time.

    Parameters
    ----------
    propensity_func : function
        Function with call signature
        `propensity_func(propensities, population, t, *args) that takes
        the current propensities and population of particle counts and
        updates the propensities for each reaction. It does not return
        anything.
    update : ndarray, shape (num_reactions, num_chemical_species)
        Entry i, j gives the change in particle counts of species j
        for chemical reaction i.
    population_0 : array_like, shape (num_chemical_species)
        Array of initial populations of all chemical species.
    time_points : array_like, shape (num_time_points,)
        Array of points in time for which to sample the probability
        distribution.
    size : int, default 1
        Number of trajectories to sample per thread.
    args : tuple, default ()
        The set of parameters to be passed to propensity_func.
    n_threads : int, default 1
        Number of threads to use in the calculation.
    progress_bar : str or bool, default False
        If True, use standard tqdm progress bar. If 'notebook', use
        tqdm.notebook progress bar. If False, no progress bar.

    Returns
    -------
    if `return_time_points` is False:
        samples : ndarray
            Entry i, j, k is the count of chemical species k at time
            time_points[j] for trajectory i. The shape of the array is
            (size*n_threads, num_time_points, num_chemical_species).
    if `return_time_points` is True:
        samples : list of 2d Numpy arrays
            samples[i][j,k] is the count of chemical species k at the
            jth time point for trajectory i.
        times : list of 1d Numpy arrays
            Entry i corresponds to a trajectory. times[i][j] is the time
            for the transition that brought the simulation to a count
            given by samples[i][j,:].
    """
    # Check inputs
    if type(args) != tuple:
        raise RuntimeError("`args` must be a tuple, not " + str(type(args)))
    population_0 = population_0.astype(int)
    update = update.astype(int)
    time_points = np.array(time_points, dtype=float)

    if len(time_points) == 2 and not return_time_points:
        warnings.warn(
            "`return_time_points` is False, and you only have two time points inputted."
        )

    if n_threads == 1:
        pop, time = _gillespie_ssa(
            propensity_func,
            update,
            population_0,
            time_points,
            return_time_points=return_time_points,
            size=size,
            args=args,
            progress_bar=progress_bar,
        )

        if return_time_points:
            return pop, time
        else:
            if len(pop) == 1:
                return pop[0].reshape((1, *pop[0].shape))
            else:
                return np.stack(pop, axis=0)
    else:
        input_args = (
            propensity_func,
            update,
            population_0,
            time_points,
            return_time_points,
            size,
            args,
            progress_bar,
        )

        with multiprocess.Pool(n_threads) as p:
            results = p.map(_gillespie_multi_fn, [input_args] * n_threads)

        pops = [results[i][0][k] for i in range(n_threads) for k in range(size)]

        if return_time_points:
            times = [results[i][1][k] for i in range(n_threads) for k in range(size)]

            return pops, times
        else:
            if len(pops) == 1:
                return pops[0].reshape((1, *pops[0].shape))
            else:
                return np.stack(pops, axis=0)
